




from ibmdbpy4nps import IdaDataBase, IdaDataFrame

from ibmdbpy4nps.ae import NZFunTApply, NZClassTApply
from ibmdbpy4nps.ae import NZFunGroupedApply

idadb = IdaDataBase('weather', 'admin', 'password')
print(idadb)

idadf = IdaDataFrame(idadb, 'KDDCUP99')
print(idadf.head())



def decision_tree_ml_host(self, df):

    from sklearn.model_selection import cross_val_score
    from sklearn.impute import SimpleImputer
    from sklearn.tree import DecisionTreeClassifier

    from sklearn.preprocessing import LabelEncoder
    import numpy as np


    result = df.groupby('PROTOCOL_TYPE')


    for name, group in result:
        # print(name)

        # print(group)
        def decision_tree_classifier(df):
            imputed_df = df.copy()



            # remove columns which have only null values
            columns = imputed_df.columns
            for column in columns:
                if imputed_df[column].isnull().sum() == len(imputed_df):
                    imputed_df = imputed_df.drop(column, 1)

            columns = imputed_df.columns

            for column in columns:

                if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                    imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                    imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))

                if (imputed_df[column].dtype == 'object'):
                    # impute missing values for categorical variables
                    imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                    imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                    imputed_df[column] = imputed_df[column].astype('str')
                    le = LabelEncoder()
                    # print(imputed_df[column].unique())

                    le.fit(imputed_df[column].unique())
                    # print(le.classes_)
                    imputed_df[column] = le.transform(imputed_df[column])

            X = imputed_df.drop('ATTACK_TYPE', axis=1)
            y = imputed_df['ATTACK_TYPE']

            # Create a decision tree
            dt = DecisionTreeClassifier(max_depth=5)

            cvscores_3 = cross_val_score(dt, X, y, cv=3)



            self.output([len(imputed_df), name, np.mean(cvscores_3)])

        ml_result = decision_tree_classifier(df=group)




output_signature = {'dataset_size':'int', 'protocol_type':'str', 'classifier_accuracy':'float'}

import time
start = time.time()
nz_tapply = NZFunTApply(df=idadf, fun_ref =decision_tree_ml_host,  parallel=False,  output_signature=output_signature)
result = nz_tapply.get_result()
result = result.as_dataframe()
print("Host only execution")
print(result)
print("\n")
end = time.time()
print(end - start)

def decision_tree_ml(self, df):
    from sklearn.model_selection import cross_val_score
    from sklearn.impute import SimpleImputer
    from sklearn.tree import DecisionTreeClassifier

    from sklearn.preprocessing import LabelEncoder
    import numpy as np

    location = df.PROTOCOL_TYPE[0]

    # data preparation
    imputed_df = df.copy()
    ds_size = len(imputed_df)



    #remove columns which have only null values
    columns = imputed_df.columns
    for column in columns:
        if imputed_df[column].isnull().sum()==len(imputed_df):
            imputed_df=imputed_df.drop(column, 1)

    columns = imputed_df.columns

    for column in columns:

        if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
            imp = SimpleImputer(missing_values=np.nan, strategy='mean')
            imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))

        if (imputed_df[column].dtype == 'object'):
            # impute missing values for categorical variables
            imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
            imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
            imputed_df[column] = imputed_df[column].astype('str')
            le = LabelEncoder()
            #print(imputed_df[column].unique())

            le.fit(imputed_df[column].unique())
            # print(le.classes_)
            imputed_df[column] = le.transform(imputed_df[column])



    X = imputed_df.drop('ATTACK_TYPE', axis=1)
    y = imputed_df['ATTACK_TYPE']

    # Create a decision tree
    dt = DecisionTreeClassifier(max_depth=5)

    cvscores_3 = cross_val_score(dt, X, y, cv=3)

    self.output(ds_size, location, np.mean(cvscores_3))




import time
start = time.time()
nz_groupapply = NZFunGroupedApply(df=idadf, index='PROTOCOL_TYPE', fun_ref=decision_tree_ml,  output_signature=output_signature)
result = nz_groupapply.get_result()
result = result.as_dataframe()
print("Parallel Execution on SPUs: \n")
print(result)
end = time.time()
print(end - start)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

"""
In-Database User defined functions
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

import inspect
import textwrap
from ibmdbpy4nps import IdaDataFrame
from builtins import dict
from future import standard_library

standard_library.install_aliases()


class NZClassTApply(object):

    def __init__(self, df, fun, input_class,  parallel,  output_signature, output_table=None):
        """
        Constructor for NZClassTApply
        """
        self.db_name = df.internal_state.current_state
        self.df = df
        self.fun = fun
        self.db = df._idadb
        self.input_class = input_class



        self.output_table = output_table
        self.output_signature = output_signature
        self.parallel =parallel

        # convert the columns index object to a list
        self.columns = self.df.columns.tolist()

    def get_result(self):
        # we need a comma separated string of column values
        columns_string = ",".join(self.columns)
        # print(columns_string)

        # get the default ae class coupled with client ml code
        # code_string = self.build_udtf_ae_code(self.fun, self.columns)
        code_string = self.build_udtf_shaper_ae_code(self.fun, self.columns, self.output_signature)

        # path_to_execute = "/nz/export/ae/adapters/python3x/11/aes/test.py"
        # code_string = "print(4+9)"

        # send the code as dynamic variable to ae function
        columns_string = columns_string + ",'CODE_TO_EXECUTE=" + "\"" + code_string + "\"" + "'"

        print("db name is " + self.db_name)
        # print(columns_string)
        if not self.parallel:
            ae_name = "py_udtf_host"
        else:
            ae_name ="py_udtf_any"
        query = "select ae_output.* from " + \
                " (select * from " + self.db_name + ") as input_t" + \
                ", table with final (" + ae_name + "(" + columns_string + ")) as ae_output"

        # print(query)
        if self.output_table:
            if self.db.exists_table(self.output_table):
              create_string = "insert into "+self.output_table+" "
            else:
              create_string = "create table " + self.output_table + " as "
            query = create_string+query
            result = self.df.ida_query(query)
            idadf = IdaDataFrame(self.db, self.output_table)
            df = idadf.as_dataframe()
            return df
        result = self.df.ida_query(query)
        return result






    def build_udtf_shaper_ae_code(self, fun, columns, output_signature):
        # we need extra single quotes for correct escaping

        input_class = self.input_class
        input_class_name = input_class.__name__
        input_class_code = inspect.getsource(input_class)



        input_class_code=input_class_code.replace(input_class_name, input_class_name+"(nzae.Ae)")
        input_class_code = input_class_code.replace("'", "''")





        import_string = """ import nzae
                            import pandas as pd
                            """
        import_string = inspect.cleandoc(import_string)

        base_code = self.get_base_shaper(columns, fun, output_signature)


        run_string = input_class_name+".run()"
        run_string = inspect.cleandoc(textwrap.dedent(run_string))

        final_code = import_string +"\n"+input_class_code+"\n"+base_code+"\n"+run_string

        #final_code = base_code + "\n" + textwrap.indent(fun_code, '     ') \
                    # + run_string


        print_string = """
        print(3+4)

        """

        return inspect.cleandoc(final_code)



    def get_base_shaper(self, columns, fun, output_signature):
        fun_name = fun.__name__
        output_signature_str = ""
        print(len(output_signature))
        for i in range(len(output_signature)):
            column_signature = output_signature[i]
            col_sig_list = column_signature.split('=')
            if col_sig_list[1]=='int':
                col_sig_list[1]= 'self.DATA_TYPE__INT32'
                output_signature_str += """
                            self.addOutputColumn('""" + col_sig_list[0] + """',""" + col_sig_list[
                    1] + """) """
            if col_sig_list[1] == 'float':
                col_sig_list[1] = 'self.DATA_TYPE__FLOAT'
                output_signature_str += """
                            self.addOutputColumn('""" + col_sig_list[0] + """',""" + col_sig_list[
                    1] + """) """
            if col_sig_list[1] == 'double':
                col_sig_list[1] = 'self.DATA_TYPE__DOUBLE'
                output_signature_str += """
                            self.addOutputColumn('""" + col_sig_list[0] + """',""" + \
                                        col_sig_list[
                                            1] + """) """
            if col_sig_list[1] == 'str':
                col_sig_list[1] = 'self.DATA_TYPE__VARIABLE'
                output_signature_str += """
                            self.addOutputColumnString('""" + col_sig_list[
                    0] + """',""" + \
                                        col_sig_list[
                                            1] + """,100) """




        #output_signature_final = inspect.cleandoc(output_signature_str)
        code_string ="""def _runUdtf(self):
                               rows_list=[]
                               for row in self:
                                 rows_list.append(row)
                               df = pd.DataFrame(rows_list, columns=""" + str(columns) + """ )
                               self.""" + fun_name + """(df)
                               
                        def _runShaper(self):
                               """+textwrap.indent(output_signature_str, '           ') + """
                                                          
        
                             
                               
                               """

        code_string = code_string.replace("'", "''")
        return inspect.cleandoc(code_string)




# -*- coding: utf-8 -*-
"""
Created on Mon Nov 23 09:53:19 2015

@author: efouche
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import dict
from future import standard_library
standard_library.install_aliases()

from collections import OrderedDict

import ibmdbpy4nps

from ibmdbpy4nps.feature_selection import entropy
from ibmdbpy4nps.feature_selection.private import _check_input

from ibmdbpy4nps.internals import idadf_state
from ibmdbpy4nps.utils import timed

import pandas as pd
import numpy as np

@idadf_state
@timed
def su(idadf, target = None, features = None, ignore_indexer=True):
    """
    Compute the symmetric uncertainty coefficients between a set of features
    and a set of target in an IdaDataFrame. 
    
    Parameters
    ----------
    idadf : IdaDataFrame
    
    target : str or list of str, optional
        A column or list of columns against to be used as target. Per default, 
        consider all columns
    
    features : str or list of str, optional
        A column or list of columns to be used as features. Per default, 
        consider all columns. 
    
    ignore_indexer : bool, default: True
        Per default, ignore the column declared as indexer in idadf
        
    Returns
    -------
    Pandas.DataFrame or Pandas.Series if only one target
    
    Notes
    -----
    Input columns as target and features should be categorical, otherwise 
    this measure does not make much sense. 
    
    Examples
    --------
    >>> idadf = IdaDataFrame(idadb, "IRIS")
    >>> su(idadf)
    """
    # Check input
    target, features = _check_input(idadf, target, features, ignore_indexer)
                
    entropy_dict = dict()
    length = len(idadf)
    corrector = np.log(length)*length
    values = OrderedDict()
        
    for t in target:
        if t not in values:
            values[t] = OrderedDict() 
        features_notarget = [x for x in features if (x != t)]
        
        for feature in features_notarget:
            if feature not in values:
                values[feature] = OrderedDict()
            if t not in values[feature]:
                if t not in entropy_dict:
                    entropy_dict[t] = entropy(idadf, t, mode = "raw")
                if feature not in entropy_dict:
                    entropy_dict[feature] = entropy(idadf, feature, mode = "raw")
                join_entropy = entropy(idadf, [t] + [feature], mode = "raw")     
                disjoin_entropy = entropy_dict[t] + entropy_dict[feature]
                value = (2.0*(disjoin_entropy - join_entropy + corrector)/(disjoin_entropy + corrector*2))
                values[t][feature] = value
                if feature in target:
                    values[feature][t] = value
    
    result = pd.DataFrame(values).fillna(np.nan)
    result = result.dropna(axis=1, how="all")
        
    if len(result.columns) > 1:
        order = [x for x in result.columns if x in features] + [x for x in features if x not in result.columns]
        result = result.reindex(order)
    
    if len(result.columns) == 1:
        if len(result) == 1:
            result = result.iloc[0,0]
        else:
            result = result[result.columns[0]].copy()
            result.sort_values(ascending = True)
    else:
        result = result.fillna(1)
   
    return result

@idadf_state
@timed
def outer_su(idadf1, key1, idadf2, key2, target = None, features1 = None, features2 = None):
    """
    Compute the symmetric uncertainty coefficients between a set of features
    and a set of target from two different IdaDataFrames on a particular key. 
    
    This is experimental 
    """
    target1, features1 = _check_input(idadf1, target, features1)
    target2, features2 = _check_input(idadf2, None, features2)
    
    if key1 not in idadf1.columns:
        raise ValueError("%s is not a column in idadf1")
    if key2 not in idadf2.columns:
        raise ValueError("%s is not a column in idadf2")
       
    condition = "a.\"%s\" = b.\"%s\""%(key1,key2)
    
    if key2 in features2:
        features2.remove(key2)
    
    afeaturesas = ", ".join(["a.\"%s\" as \"a.%s\" "%(feature, feature) for feature in features1])
    bfeaturesas = ", ".join(["b.\"%s\" as \"b.%s\" "%(feature, feature) for feature in features2])
    
    selectlist = [afeaturesas, bfeaturesas]
    
    if target1 is not None:
        atargetas = ", ".join(["a.\"%s\" as \"a.%s\" "%(tar, tar) for tar in [target1]])
        selectlist.append(atargetas)
        atarget = "a." + target1
    else:
        atarget = None
        
    abfeatures = ["a." + feature for feature in features1] + ["b." + feature for feature in features2]
    selectstr = ", ".join(selectlist)
    
    expression = "SELECT %s FROM %s as a FULL OUTER JOIN %s as b ON %s"%(selectstr, idadf1.name, idadf2.name, condition)
    
    viewname = idadf1._idadb._create_view_from_expression(expression)
    
    try:
        idadf_join = ibmdbpy4nps.IdaDataFrame(idadf1._idadb, viewname)
        return su(idadf_join, target = atarget, features = abfeatures)
    except:
        raise
    finally:
        idadf1._idadb.drop_view(viewname)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

# Python 2 Compatibility
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import dict
from builtins import zip
from builtins import str
from builtins import int
from future import standard_library
standard_library.install_aliases()

from collections import OrderedDict
import itertools
import math
import warnings
from numbers import Number

import pandas as pd
import numpy as np
import six

import ibmdbpy4nps
from ibmdbpy4nps.utils import chunklist

"""
Statistics module for IdaDataFrames
"""

def _numeric_stats(idadf, stat, columns):
    """
    Compute various stats from one or several numerical columns of an IdaDataFrame.

    Parameters
    ----------
    idadf : IdaDataFrame
        Data source.
    stat : str
        Name of the statistic to be computed.
    columns : str or list of str
        Name of the columns that belong to the IdaDataFrame.

    Returns
    -------
    Tuple or float64
        One value for each column. For a one column input a float64 value is returned except for median

    Notes
    -----
    Currently, the following functions are supported: count, mean, median, std,
    var, min, max, sum. Should return a tuple. Only available for numerical
    columns.
    """
    # Calculate count, mean, median, std, var, min, max
    if isinstance(columns, six.string_types):
        columns = [columns]

    if isinstance(stat, six.string_types):
        if stat == "count":
            select_string = 'COUNT(\"' + '\"), COUNT(\"'.join(columns) + '\")'
        elif stat == "mean":
            select_string = ('AVG(CAST(\"' +
                             '\" AS FLOAT)), AVG(CAST(\"'.join(columns) +
                             '\" AS FLOAT))')
        elif stat == "median":
            return _get_percentiles(idadf, 0.5, columns).values[0]
        elif stat == "std":
            tuple_count = _numeric_stats(idadf, 'count', columns)
            # in case of only one column, ensure tuple_count is iterable
            if len(columns) == 1:
                tuple_count = [tuple_count]
            count_dict = dict((x, y) for x, y in zip(columns, tuple_count))
            agg_list = []
            for column in columns:
                if idadf._idadb._is_netezza_system():
                    # workaround for error on Netezza related to "/"
                    # ERROR:  Function 'QRT(INT4)' does not exist
                    # sqrt_term = "SQRT(%s) * POW(SQRT(%s), -1)" %(count_dict[column], count_dict[column]-1)
                    sqrt_term = 1
                else:
                    sqrt_term = "SQRT(%s)/SQRT(%s)"  %(count_dict[column], count_dict[column]-1)
                # agg_list.append("STDDEV(\"%s\")*(SQRT(%s)/SQRT(%s))"
                agg_list.append("STDDEV(\"%s\")*( %s )" %(column, sqrt_term))
            select_string = ', '.join(agg_list)
        elif stat == "var":
            tuple_count = _numeric_stats(idadf, 'count', columns)
            if len(columns) == 1:
                tuple_count = [tuple_count]
            count_dict = dict((x, int(y)) for x, y in zip(columns, tuple_count))
            agg_list = []
            for column in columns:
                if idadf._idadb._is_netezza_system():
                # workaround for error on Netezza related to "/"
                # ERROR:  Unable to identify an operator '//' ..
                    # div_term = "%s.0 * POW(%s.0, -1)" %(count_dict[column], count_dict[column]-1)
                    div_term = 1
                else:
                    div_term = "%s.0/%s.0" %(count_dict[column], count_dict[column]-1)
                agg_list.append("VARIANCE(\"%s\")*(%s)" %(column, div_term))
            select_string = ', '.join(agg_list)
        elif stat == "min":
            select_string = 'MIN(\"' + '\"), MIN(\"'.join(columns) + '\")'
        elif stat == "max":
            select_string = 'MAX(\"' + '\"), MAX(\"'.join(columns) + '\")'
        elif stat == "sum":
            select_string = 'SUM(\"' + '\"), SUM(\"'.join(columns) + '\")'

        name = idadf.internal_state.current_state

        return idadf.ida_query("SELECT %s FROM %s" %(select_string, name)).values[0]


def _get_percentiles(idadf, percentiles, columns):
    """
    Return percentiles over all entries of a column or list of columns in the
    IdaDataFrame.

    Parameters
    ----------
    idadf : IdaDataFrame
    percentiles: Float or list of floats.
        All values in percentiles must be > 0  and < 1
    columns: String or list of string
        Name of columns belonging to the IdaDataFrame.

    Returns
    -------
        DataFrame
    """

    if isinstance(columns, six.string_types):
        columns = [columns]
    if isinstance(percentiles, Number):
        percentiles = [percentiles]

    name = idadf.internal_state.current_state

    # Get na values for each columns
    tuple_na = _get_number_of_nas(idadf, columns)
    nrow = idadf.shape[0]
    data = pd.DataFrame()
    for index_col, column in enumerate(columns):
        nb_not_missing = nrow - tuple_na[index_col]
        indexes = [float(x)*float(nb_not_missing-1) + 1 for x in percentiles]
        low = [math.floor(x) for x in indexes]
        high = [math.ceil(x) for x in indexes]
        tuplelist = []
        i = 0
        for flag in [((x+1) == y) for x, y in zip(low, high)]:
            if flag:
                tuplelist.append((i, i+1))
                i += 2
            else:
                tuplelist.append((i, i))
                i += 1
        unique = low + high
        unique = set(unique)
        unique = sorted(unique)
        unique = [str(x) for x in unique]
        indexes_string = ",".join(unique)
        df = idadf.ida_query("(SELECT \""+column+"\" AS \""+column+"\" FROM (SELECT "+
                        "ROW_NUMBER() OVER(ORDER BY \""+column+"\") as rn, \""+
                        column + "\" FROM (SELECT * FROM " + name +
                        " WHERE " + column + " IS NOT NULL " +
                        ") AS T1) AS T2 WHERE rn  in("+ indexes_string +"))")

        #indexvalues = list(df[df.columns[0]])
        indexvalues = list(df)
        #import pdb ; pdb.set_trace()
        #print(tuplelist)
        #print(indexvalues)
        indexfinal = [(float(str(indexvalues[x[0]]))+float(str(indexvalues[x[1]])))/2 for x in tuplelist]
        new_data = pd.DataFrame(indexfinal)
        data[column] = (new_data.T).values[0]

    percentile_names = [x for x in percentiles]
    data.index = percentile_names
    return data


def _categorical_stats(idadf, stat, columns):
    # TODO:
    """
    Computes various stats from one or several categorical columns of the IdaDataFrame.
    This is not implemented.

    Parameters
    ----------
    idadf : IdaDataFrame
    stat : str
        Name of the statistic function to be computed.
    columns : str or list of str
        Name of columns belonging to the IdaDataFrame.

    Returns
    -------
        Tuple.
    """
    # Calculates count, unique, top, freq
    raise NotImplementedError("TODO")

def _get_number_of_nas(idadf, columns):
    """
    Return the count of missing values for a list of columns in the IdaDataFrame.

    Parameters
    ----------
    idadf : IdaDataFrame
    columns : str or list
        One column as a string or a list of columns in the idaDataFrame.

    Returns
    -------
        Tuple
    """
    if isinstance(columns, six.string_types):
        columns = [columns]

    name = idadf.internal_state.current_state

    query_list = list()
    for column in columns:
        string = ("(SELECT COUNT(*) AS \"" + column + "\" FROM " +
                name + " WHERE \"" + column + "\" IS NULL) AS T_" + column)
        query_list.append(string)

    query_string = ', '.join(query_list)

    # TODO: Improvement idea : Get nrow (shape) and substract by count("COLUMN")
    return idadf.ida_query("SELECT * FROM " + query_string, first_row_only = True)

def _count_level(idadf, columnlist=None):
    """
    Count distinct levels across a list of columns of an IdaDataFrame grouped
    by themselves.


    Parameters
    ----------
    columnlist : list
        List of column names that exist in the IdaDataFrame. By default, these
        are all columns in IdaDataFrame.

    Returns
    -------
        Tuple

    Notes
    -----
    The function assumes the following:

        * The columns given as parameter exists in the IdaDataframe.
        * The parameter columnlist is an optional list.
        * Columns are referenced by their own name (character string).
    """
    if columnlist is None:
        columnlist = idadf.columns

    name = idadf.internal_state.current_state

    query_list = []
    for column in columnlist:
        # Here cast ?
        query_list.append("(SELECT COUNT(*) AS \"" + column +"\" FROM (" +
                          "SELECT \"" + column + "\" FROM " + name +
                          " GROUP BY \"" + column + "\" ) AS T1_" + column + ") AS T2_" + column)
        #query_list.append("(SELECT CAST(COUNT(*) AS BIGINT) AS \"" + column +"\" FROM (" +
        #                  "SELECT \"" + column + "\" FROM " + name + " ))")

    query_string = ', '.join(query_list)
    column_string = '\"' + '\", \"'.join(columnlist) + '\"'
    return idadf.ida_query("SELECT " + column_string + " FROM " + query_string, first_row_only = True)

def _count_level_groupby(idadf, columnlist=None):
    """
    Count distinct levels across a list of columns in the IdaDataFrame grouped
    by themselves. This is used to get the dimension of the resulting cross table.

    Parameters
    ----------
    columnlist : list
        List of column names existing in the IdaDataFrame. By default, these
        are columns of self

    Returns
    -------
        Tuple

    Notes
    -----
    The function assumes the follwing:

        * The columns given as parameter exists in the IdaDataframe.
        * The parameter columnlist is a optional and is a list.
        * Columns are referenced by their own name (character string).
    """
    if columnlist is None:
        columnlist = idadf.columns

    name = idadf.internal_state.current_state

    column_string = '\"' + '\", \"'.join(columnlist) + '\"'
    query = (("SELECT COUNT(*) FROM (SELECT %s, COUNT(*) as COUNT "+
            "FROM %s GROUP BY %s ORDER BY %s, COUNT ASC) AS T ")
            %(column_string, name, column_string, column_string))
    return idadf.ida_query(query, first_row_only = True)

# TODO: REFACTORING: factors function should maybe return a tuple ?
def _factors_count(idadf, columnlist, valuelist=None):
    """
    Count non-missing values for all columns in a list (valuelist) over the
    IdaDataFrame grouped by a list of columns(columnlist).

    Parameters
    ----------
    columnlist : list
        List of column names that exist in self.
    valuelist : list
        List of column names that exist in self.

     Notes
     -----
     The function assumes the following:

        * The columns given as parameter exists in the IdaDataframe
        * The parameter columnlist is a optional and is a list
        * Columns are referenced by their own name (character string)

    Returns
    -------
        DataFrame
    """
    column_string = '\"' + '\", \"'.join(columnlist) + '\"'

    name = idadf.internal_state.current_state

    if valuelist is None:
        query = (("SELECT %s, COUNT(*) as COUNT FROM %s GROUP BY %s ORDER BY %s, COUNT ASC")
                %(column_string, name, column_string, column_string))
    else:
        agg_list = []
        for value in valuelist:
            query = "COUNT(\"%s\") as \"%s\""%(value,value)
            agg_list.append(query)

        agg_string =  ', '.join(agg_list)
        value_string = '\"' + '", "'.join(valuelist) + '\"'

        query = (("SELECT %s,%s FROM %s GROUP BY %s ORDER BY %s,%s ASC")
                %(column_string, agg_string, name, column_string, column_string, value_string))

    return idadf.ida_query(query)

def _factors_sum(idadf, columnlist, valuelist):
    """
    Compute the arithmetic sum over for all columns in a list (valuelist)
    over the IdaDataFrame grouped by a list of columns (columnlist).

    Parameters
    ----------
    columnlist : list
        List of column names that exist in self.
    valuelist : list
        List of column names that exist in self.

    Notes
    -----
    The function assumes the following:

        * The columns given as parameter exists in the IdaDataframe
        * The parameter columnlist is a optional and is a list
        * Columns are referenced by their own name (character string)

    Returns
    -------
        DataFrame
    """
    column_string = '\"' + '\", \"'.join(columnlist) + '\"'

    name = idadf.internal_state.current_state

    agg_list = []
    for value in valuelist:
        query = "SUM(\"%s\") as \"%s\""%(value, value)
        agg_list.append(query)

    agg_string =  ', '.join(agg_list)
    value_string = '\"' + '", "'.join(valuelist) + '\"'

    query = (("SELECT %s,%s FROM %s GROUP BY %s ORDER BY %s,%s ASC")
            %(column_string, agg_string, name, column_string, column_string, value_string))

    return idadf.ida_query(query)

def _factors_avg(idadf, columnlist, valuelist):
    """
    Compute the arithmetic average for all columns in a list (valuelist) over
    the IdaDataFrame grouped by a list of columns (columnlist).

    Parameters
    ----------
    columnlist : list
        List of column names that exist in self.
    valuelist : list
        List of column names that exist in self.

    Notes
    -----
    The function assumes the following:

        * The columns given as parameter exists in the IdaDataframe
        * The parameter columnlist and valuelist are array-like
        * Columns are referenced by their own name (character string)

    Returns
    -------
        DataFrame
    """
    column_string = '\"' + '\", \"'.join(columnlist) + '\"'

    name = idadf.internal_state.current_state

    agg_list = []
    for value in valuelist:
        agg = (("CAST(AVG(CAST(\"%s\" AS DECIMAL(10,6))) AS DECIMAL(10,6)) \"%s\"")
              %(value, value))
        agg_list.append(agg)


    agg_string =  ', '.join(agg_list)
    value_string = '\"' + '", "'.join(valuelist) + '\"'

    query = (("SELECT %s,%s FROM %s GROUP BY %s ORDER BY %s,%s ASC")
            %(column_string, agg_string, name, column_string, column_string, value_string))

    return idadf.ida_query(query)

###############################################################################
### Pivot Table
###############################################################################

def pivot_table(idadf, values=None, columns=None, max_entries=1000, sort=None,
                factor_threshold=None, interactive=False, aggfunc='count'):
    """
    See IdaDataFrame.pivot_table
    """

    # TODO : Support index

    if aggfunc.lower() not in ['count', 'sum', 'avg', 'average', 'mean']:
        print("For now only 'count' and 'sum' and 'mean' as aggregation function is supported")
        return

    if (columns is None) & (factor_threshold is None):
        print("Please provide parameter factor_threshold for automatic selection of columns")
        return

    if isinstance(columns, six.string_types):
        columns = [columns]

    if isinstance(values, six.string_types):
        values = [values]

    if (values is None) and (aggfunc.lower() != "count"):
        raise ValueError("Cannot aggregate using another function than count if" +
                         "no value(s) was/were given")

    ####### Identify automatically categorical fields #########
    # Load distinct count for each and evaluate categorical or not
    data = idadf._table_def(factor_threshold) #
    if columns is None:
        factors = data.loc[data['VALTYPE'] == "CATEGORICAL", ['TYPENAME', 'FACTORS']]
        if len(factors) == 0:
            print("No categorical columns to tabulate")
            return
    else:
        factors = data.loc[columns, ['TYPENAME', 'FACTORS']]

    if sort == "alpha":
        factors.sort_index(inplace=True, ascending=1)
    elif sort == "factor":
        factors.sort(['FACTORS'], inplace=True, ascending=1)

    if columns is None:
        print("Automatic selection of columns :", factors.index.values)
        columns = factors.index.values

    nb_row = _count_level_groupby(idadf, factors.index.values)[0] * len(columns)
    nb_col = len(factors.index.values)

    nb_entries = nb_row * nb_col

    if nb_entries > max_entries: # Overflow risk
        print("Number of entries :", nb_entries)
        print("Value counts for factors:")
        factor_values = factors[['FACTORS']]
        factor_values.columns = ['']
        print(factor_values.T)
        print("WARNING :Attempt to make a table with more than " +
              str(max_entries)+ " elements. Either increase max_entries " +
              "parameter or remove columns with too many levels.")
        return

    print("Output dataframe has dimensions", nb_row, "x", (nb_col+1))
    if interactive is True:
        display_yes = ibmdbpy4nps.utils.query_yes_no("Do you want to download it in memory ?")
        if not display_yes:
            return

    categorical_columns = list(factors.index)
    if aggfunc.lower() == 'count':
        dataframe = _factors_count(idadf, categorical_columns, values) # Download dataframe
    if aggfunc.lower() == 'sum':
        dataframe = _factors_sum(idadf, categorical_columns, values) # Download dataframe
    if aggfunc.lower() in ['avg', 'average', 'mean']:
        dataframe = _factors_avg(idadf, categorical_columns, values) # Download dataframe

    if values is not None:
        agg_values = values
    else: agg_values = aggfunc.upper()

    if isinstance(agg_values, six.string_types):
        agg_values = [agg_values]
    dataframe.columns = categorical_columns + agg_values # Name the aggregate column

    # Formatting result
    if len(agg_values) == 1:
        dataframe[None] = agg_values[0]
    else:
        catdataframe = dataframe[categorical_columns]
        dataframe = catdataframe.join(dataframe[agg_values].stack().reset_index(1))
        dataframe['level_1'] = pd.Categorical(dataframe['level_1'], agg_values)
        dataframe = dataframe.rename(columns={'level_1':None})
        dataframe = dataframe.sort([None] + categorical_columns)

    dataframe.set_index([None] + categorical_columns, inplace=True)
    dataframe = dataframe.astype(float)

    result = pd.Series(dataframe[dataframe.columns[0]])
    result.name = None

    return result

###############################################################################
### Descriptive statistics
###############################################################################

def summary(idadf):

        table_name= idadf.internal_state.current_state
        outtable_name = idadf._idadb._get_valid_tablename(prefix="pyida_describe")
        idadf._idadb._call_stored_procedure("SUMMARY1000 ", intable=table_name, outtable=outtable_name)
        result_query = "SELECT * FROM "+outtable_name+" ORDER BY columnname; "
        result_df = idadf.ida_query(result_query)
        idadf._idadb._call_stored_procedure("DROP_SUMMARY1000", intable=outtable_name)
        return result_df

def describe(idadf, percentiles=[0.25, 0.50, 0.75]):
    """
    See IdaDataFrame.describe
    """
    if percentiles is not None:
        if isinstance(percentiles, Number):
            percentiles = [percentiles]
        if True in [(not isinstance(x, Number)) for x in percentiles]:
            raise TypeError("Argument 'percentiles' should be either a number or " +
                            "a list of numbers between 0 and 1")
        elif True in [((x >= 1) | (x <= 0)) for x in percentiles]:
            raise ValueError("Numbers in argument 'percentiles' should be between 0 and 1")

    # Improvement idea : We could use dtypes instead of calculating this everytime
    columns = idadf._get_numerical_columns()
    data = []
    if not columns:
        columns = idadf._get_categorical_columns()
        if not columns:
            raise NotImplementedError("No numerical and no categorical columns")
        else:
            raise NotImplementedError("Categorical only idaDataFrame are not handled currently")
            # TODO : Handle categorical columns
            data.append(_categorical_stats(idadf, "count", columns))
            data.append(_categorical_stats(idadf, "unique", columns))
            data.append(_categorical_stats(idadf, "top", columns))
            data.append(_categorical_stats(idadf, "freq", columns))
    else:
        data.append(_numeric_stats(idadf, "count", columns))
        data.append(_numeric_stats(idadf, "mean", columns))
        data.append(_numeric_stats(idadf, "std", columns))
        data.append(_numeric_stats(idadf, "min", columns))
        if percentiles is not None:
            perc = (_get_percentiles(idadf, percentiles, columns))
            for tup in perc.itertuples(index=False):
                data.append(tup)
        data.append(_numeric_stats(idadf, "max", columns))

    data = pd.DataFrame(data)
    data.columns = columns
    if percentiles is not None:
        percentile_names = [(str(int(x * 100)) + "%") for x in percentiles]
    else:
        percentile_names = []
    data.index = ['count', 'mean', 'std', 'min'] + percentile_names + ['max']

    # quick fix -> JDBC problems
    #for column in data.columns:
    #    data[[column]] = data[[column]].astype(float)

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        data = pd.Series(data[data.columns[0]])

    return data


def quantile(idadf, q=0.5):
    """
    See IdaDataFrame.quantile
    """

    if isinstance(q, Number):
        q = [q]

    # Sanity check
    if True in [(not isinstance(x, Number)) for x in q]:
        raise TypeError("Argument 'q' should be either a number or " +
                        "a list of numbers between 0 and 1")
    elif True in [((x >= 1) | (x <= 0)) for x in q]:
        raise ValueError("Numbers in argument 'percentiles' should be between 0 and 1")

    columns = idadf._get_numerical_columns()
    if not columns:
        print(idadf.name + " has no numeric columns")
        return

    result = _get_percentiles(idadf, q, columns)

    if isinstance(q, list):
        if len(q) > 1:
            return result

    result = result.T
    result = result[result.columns[0]]
    result.name = q[0]
    result = result.astype('float')

    if len(result) == 1:
        result = result[0]

    return result

# Note : Not casting to double can lead to SQL overflow
# TODO: Has to be modified in ibmdbR


def cov(idadf, other=None):
    numerical_columns = idadf._get_numerical_columns()
    if len(numerical_columns) < 2:
        print(idadf.name + " has less than two numeric columns")
        return
    column_string = ""
    for column in numerical_columns:
        column_string += column + ";"

    result_df = pd.DataFrame(columns=numerical_columns, index=numerical_columns)



    table_name = idadf.internal_state.current_state
    outtable = idadf._idadb._get_valid_tablename(prefix="cov_")

    idadf._idadb._call_stored_procedure("COVARIANCE1000MATRIX",
                                        intable=table_name,
                                        incolumn=column_string,
                                        outtable=outtable)

    result_query = "SELECT * FROM " + outtable + " ORDER BY varxname, varyname; "

    cov_df = idadf.ida_query(result_query)


    for index in cov_df.index.values:

        col_list = []
        for column in cov_df.columns.values:
            col_list.append(cov_df.at[index, column])

        result_df.at[col_list[0], col_list[1]] = col_list[2]

    for column in result_df.columns:
        result_df[column] = result_df[column].astype(float)
    value = idadf._idadb.drop_table(outtable)

    return result_df


def cov_old(idadf, other = None):
    """
    See IdaDataFrame.cov
    """
    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        raise TypeError("cov() missing 1 required positional argument: 'other'")

    # check if the covariance function is installed on the Netezza system
    if idadf._idadb._is_netezza_system():
        covar_query = "SELECT count(*) from _V_OBJECT  where OBJNAME like 'COVAR_SAMP#%' AND OBJDB = CURRENT_DB"
        if idadf._idadb.ida_scalar_query(covar_query) == 0:
            raise NotImplementedError("The COVAR_SAMP function is not installed on the Netezza database.")

    columns = idadf._get_numerical_columns()
    if len(columns) < 2 :
        print(idadf.name + " has less than two numeric columns")
        return

    tuple_count = _numeric_stats(idadf, 'count', columns)
    count_dict = dict((x, int(y)) for x, y in zip(columns, tuple_count))

    agg_list = []

    combinations = [x for x in itertools.combinations_with_replacement(columns, 2)]
    columns_set = [{x[0], x[1]} for x in combinations]

    if idadf._idadb._is_netezza_system():
        for column_pair in combinations:
            agg_list.append("COVAR_SAMP(\"" + column_pair[0] + "\",\"" + column_pair[1] + "\")")
    else:
        for column_pair in combinations:
            agg_list.append("COVARIANCE(\"" + column_pair[0] + "\",\"" +
                            column_pair[1] + "\")*(" +
                            str(min([count_dict[column_pair[0]],
                                 count_dict[column_pair[1]]])) + ".0/" +
                            str(min([count_dict[column_pair[0]],
                                 count_dict[column_pair[1]]])-1) + ".0)")

    agg_string = ', '.join(agg_list)

    name = idadf.internal_state.current_state

    data = idadf.ida_query("SELECT %s FROM %s"%(agg_string, name), first_row_only = True)

    tuple_list = []

    for column1 in columns:
        list_value = []
        for column2 in columns:
            for index, column_set in enumerate(columns_set):
                if {column1, column2} == column_set:
                    list_value.append(data[index])
                    break
        tuple_list.append(tuple(list_value))

    result = pd.DataFrame(tuple_list)
    result.index = columns
    result.columns = columns

    if len(result) == 1:
        result = result[0]

    return result

def corr(idadf):
    numerical_columns = idadf._get_numerical_columns()
    if len(numerical_columns) < 2 :
        print(idadf.name + " has less than two numeric columns")
        return
    column_string=""
    for column in numerical_columns:
        column_string+=column+";"

    result_df = pd.DataFrame(columns=numerical_columns, index=numerical_columns)

    #print(result_df)

    table_name = idadf.internal_state.current_state
    outtable = idadf._idadb._get_valid_tablename(prefix="corr_")

    idadf._idadb._call_stored_procedure("CORRELATION1000MATRIX ",
                                        intable=table_name,
                                        incolumn=column_string,
                                        outtable=outtable)
    #corr_df = idadf.ida_query(create_corr_query)
    result_query = "SELECT * FROM " + outtable+ " ORDER BY varxname, varyname; "



    corr_df = idadf.ida_query(result_query)

    for index in corr_df.index.values:

        col_list = []
        for column in corr_df.columns.values:

            col_list.append(corr_df.at[index, column])


        result_df.at[col_list[0], col_list[1]] = col_list[2]

    for column in result_df.columns:

        result_df[column] = result_df[column].astype(float)
    value = idadf._idadb.drop_table(outtable)
    return result_df


def corr_old(idadf, features=None,ignore_indexer=True):
    """
    See IdaDataFrame.corr
    """
    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        raise TypeError("corr() missing 1 required positional argument: 'other'")

    # check if the corr function is installed on the Netezza system
    if idadf._idadb._is_netezza_system():
        corr_query = "SELECT count(*) from _V_OBJECT  where OBJNAME like 'CORR#%' AND OBJDB = CURRENT_DB"
        if idadf._idadb.ida_scalar_query(corr_query) == 0:
            raise NotImplementedError("The CORR function is not installed on the Netezza database.")

    numerical_columns = idadf._get_numerical_columns()
    if len(numerical_columns) < 2 :
        print(idadf.name + " has less than two numeric columns")
        return

    if ignore_indexer is True:
        if idadf.indexer:
            if idadf.indexer in numerical_columns:
                numerical_columns.remove(idadf.indexer)

    #print(features)
    #target, features = ibmdbpy.utils._check_input(target, features)
    if features is not None:
        for feature in features:
            if feature not in numerical_columns:
                raise TypeError("Correlation-based measure not available for non-numerical columns %s"%feature)
    else:
        features = numerical_columns

    #if target not in columns:
    #    raise ValueError("%s is not a column of numerical type in %s"%(target, idadf.name))

    values = OrderedDict()

    combinations = [x for x in itertools.combinations(features, 2)]
    #columns_set = [{x[0], x[1]} for x in combinations]

    if idadf._idadb._is_netezza_system():
        corr_function = "CORR"
    else:
        corr_function = "CORRELATION"

    if len(features) < 64: # the limit of variables for an SQL statement is 4096, i.e 64^2
        agg_list = []
        for column_pair in combinations:
            agg = corr_function + "(\"%s\",\"%s\")"%(column_pair[0], column_pair[1])
            agg_list.append(agg)

        agg_string = ', '.join(agg_list)

        name = idadf.internal_state.current_state

        data = idadf.ida_query("SELECT %s FROM %s"%(agg_string, name), first_row_only = True)

        for i, element in enumerate(combinations):
            if element[0] not in values:
                values[element[0]] = {}
            if element[1] not in values:
                values[element[1]] = {}
            values[element[0]][element[1]] = data[i]
            values[element[1]][element[0]] = data[i]

        result = pd.DataFrame(values).fillna(1)
    else:
        chunkgen = chunklist(combinations, 100)

        for chunk in chunkgen:
            agg_list = []
            for column_pair in chunk:
                agg = corr_function + "(\"%s\",\"%s\")"%(column_pair[0], column_pair[1])
                agg_list.append(agg)

            agg_string = ', '.join(agg_list)

            name = idadf.internal_state.current_state

            data = idadf.ida_query("SELECT %s FROM %s"%(agg_string, name), first_row_only = True)

            for i, element in enumerate(chunk):
                if element[0] not in values:
                    values[element[0]] = OrderedDict()
                if element[1] not in values:
                    values[element[1]] = OrderedDict()
                values[element[0]][element[1]] = data[i]
                values[element[1]][element[0]] = data[i]

        result = pd.DataFrame(values).fillna(1)

    result = result.reindex(result.columns)
    if len(result) == 1:
        result = result[0]

    return result

### corrwith


def mad(idadf):
    """
    See IdaDataFrame.mad
    """
    columns = idadf._get_numerical_columns()
    if len(columns) < 2 :
        print(idadf.name + " has less than two numeric columns")
        return

    mean_tuple = _numeric_stats(idadf, "mean", columns)
    absmean_dict = dict((x, abs(y)) for x, y in zip(columns, mean_tuple))
    tuple_na = _get_number_of_nas(idadf, columns)

    agg_list = []
    for index_col, column in enumerate(columns):
        if idadf._idadb._is_netezza_system():
            # workaround for error on Netezza related to "/"
            # ERROR:  Unable to identify an operator '//'
            div_term = "* pow(" + str(idadf.shape[0] - tuple_na[index_col]) + ", -1)"
        else:
            div_term = "/" + str(idadf.shape[0] - tuple_na[index_col])
        agg_list.append("SUM(ABS(\"" + column + "\" -" +
                        str(absmean_dict[column]) + "))" +
                        div_term)

    agg_string = ', '.join(agg_list)

    name = idadf.internal_state.current_state

    mad_tuple = idadf.ida_query("SELECT %s FROM %s"%(agg_string, name))
    result = pd.Series(mad_tuple.values[0])
    result.index = columns
    result = result.astype('float')

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def ida_min(idadf):
    """
    See idadataFrame.min
    """
    na_tuple = _get_number_of_nas(idadf, idadf.columns)
    min_tuple = _numeric_stats(idadf, "min", idadf.columns)
    if not hasattr(min_tuple,"__iter__") : min_tuple = (min_tuple,) # dirty fix
    min_list = [np.nan if ((y > 0) and not isinstance(x, Number))
                else x for x, y in zip(min_tuple, na_tuple)]
    min_tuple = tuple(min_list)
    result = pd.Series(min_tuple)
    result.index = idadf.columns

    #if isinstance(idadf, ibmdbpy.IdaSeries):
     #   result = result[0]

    return result

def ida_max(idadf):
    """
    See idadataFrame.max
    """
    na_tuple = _get_number_of_nas(idadf, idadf.columns)
    max_tuple = _numeric_stats(idadf, "max", idadf.columns)
    if not hasattr(max_tuple,"__iter__") : max_tuple = (max_tuple,) # dirty fix
    max_list = [np.nan if ((y > 0) and not isinstance(x, Number))
                else x for x, y in zip(max_tuple, na_tuple)]
    max_tuple = tuple(max_list)
    result = pd.Series(max_tuple)
    result.index = idadf.columns

    #if isinstance(idadf, ibmdbpy.IdaSeries):
      #  result = result[0]

    return result

def count(idadf):
    """
    See IdaDataFrame.count
    """
    count_tuple = _numeric_stats(idadf, "count", idadf.columns)
    result = pd.Series(count_tuple)
    result.index = idadf.columns
    result = result.astype(int)

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def count_distinct(idadf):
    """
    See IdaDataFrame.count_distinct
    """
    result = pd.Series(_count_level(idadf))
    result.index = idadf.columns
    result = result.astype(int)

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def std(idadf):
    """
    See IdaDataFrame.std
    """
    columns = idadf._get_numerical_columns()
    if not columns:
        warnings.warn("%s has no numeric columns"%idadf.name)
        return pd.Series()

    std_tuple = _numeric_stats(idadf, "std", columns)

    result = pd.Series(std_tuple)
    result.index = columns

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def var(idadf):
    """
    See IdaDataFrame.var
    """
    columns = idadf._get_numerical_columns()
    if not columns:
        warnings.warn("%s has no numeric columns"%idadf.name)
        return pd.Series()

    var_tuple = _numeric_stats(idadf, "var", columns)

    result = pd.Series(var_tuple)
    result.index = columns

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def mean(idadf):
    """
    See IdaDataFrame.mean
    """
    columns = idadf._get_numerical_columns()
    if not columns:
        warnings.warn("%s has no numeric columns"%idadf.name)
        return pd.Series()

    mean_tuple = _numeric_stats(idadf, "mean", columns)

    result = pd.Series(mean_tuple)
    result.index = columns

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def ida_sum(idadf):
    """
    See IdaDataFrame.sum
    """
    #Behave like having the option "numeric only" to true
    columns = idadf._get_numerical_columns()
    if not columns:
        warnings.warn("%s has no numeric columns"%idadf.name)
        return pd.Series()

    sum_tuple = _numeric_stats(idadf, "sum", columns)

    result = pd.Series(sum_tuple)
    result.index = columns

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result

def median(idadf):
    """
    See IdaDataFrame.median
    """
    #Behave like having the option "numeric only" to true
    columns = idadf._get_numerical_columns()
    if not columns:
        warnings.warn("%s has no numeric columns"%idadf.name)
        return pd.Series()

    median_tuple = _numeric_stats(idadf, "median", columns)

    result = pd.Series(median_tuple)
    result.index = columns

    if isinstance(idadf, ibmdbpy4nps.IdaSeries):
        result = result[0]

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
In-Database Association Rules Mining
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import dict
from future import standard_library
standard_library.install_aliases()

import ibmdbpy4nps
from ibmdbpy4nps.exceptions import IdaAssociationRulesError
import six

#----------------------------------------------------------------------
# AssociationRules class
class AssociationRules(object):
    """
    Association rules mining can be used to discover interesting and useful 
    relations between items in a large-scale transaction table. You can 
    identify strong rules between related items by using different measures of 
    relevance. Apriori or FP-Growth are well-known algorithms for association 
    rules mining. For analytic stored procedures, the PrefixSpan algorithm is 
    preferred due to its scalability.

    The AssociationRules class provides an interface for using the
    ASSOCRULES amd PREDICT_ASSOCRULES IDAX/INZA methods.
    """

    def __init__(self, modelname = None, minsupport = None, maxlen = 5, maxheadlen = 1, minconf = 0.5):
        """
        Constructor for association rules model

        Parameters
        ----------
        modelname : str
            The name of the Association Rules model that will be built.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If it is not given, it will be generated automatically.
            If the parameter corresponds to an existing model in the database,
            it will be replaced during the fitting step.
        
        minsupport : float or integer, optional
            The minimum fraction (0.0 - 1.0) or the minimum number (above 1) of
            transactions that must contain a pattern to be considered as frequent.

            Default:
                * system-determined

            Range:
                * >0.0 and <1.0 for a minimum fraction
                * >1 for a minimum number of transactions.
        
        maxlen : int, optional, >=2, default: 5
            The maximum length of a pattern or a rule, that is,
            the maximum number of items per pattern or rule.
        
        maxheadlen : int, optional, >= 1 and <maxlen, default: 1
            The maximum length of a rule head, that is, the maximum number of
            items that might belong to the item set on the right side of a rule.
            Increasing this value might significantly increase the number of detected rules.
        
        minconf : float, optional, >=0.0 and <= 1, default: 0.5
            The minimum confidence that a rule must achieve to be kept in the model of the pattern.
            
        Attributes
        ----------
               
        nametable
            Gets set at fit step. The name of the optional table which contains a mapping
            between the items from the input table and the item names.
            This attribute is set through the nametable option of the fit method.
            This table must contain at least two columns, where

            * The first column has the same name as the column that is
              contained in the item parameter of the input table
            * The second column has the same name as the name that is
              defined in the namecol parameter
        
        namecol
            Gets set at fit step. The name of the optional column which
            contains item names as defined in the nametable attribute.
            This attribute is set through the fit method and cannot be specified
            of the nametable parameter of fit is not specified.
        
        outtable
            Gets set at predict step, name of the optional output table
            in which the mapping between the input
            sequences and the associated rules or patterns is written.
            If the parameter corresponds to an existing table in the database,
            it is replaced.
        
        type : str, optional, default : "rules".
            Gets set at predict step.
            The type of information that is written in the output table.
            The following values are possible: ‘rules’ and ‘patterns’;
        
        limit: Gets set at predict step int, optional, >=1, default: 1
            The maximum number of rules or patterns that is written in the 
            output table for each input sequence.;
        
        sort: Gets set at predict step, str or list, optional
            A list of keywords that indicates the order in which the rules or 
            patterns are written in the output table. The order of the list is 
            descending. The items are separated by semicolons. The following 
            values are possible: ‘support’, ‘confidence’, ‘lift’, and ‘length’. 
            The ‘confidence’ value can only be specified if the type parameter 
            is ‘rules’. If the type parameter is ‘rules’, the default is: 
            support;confidence;length.  If the type parameter is ‘patterns’, 
            the default is: support;lift;length. ;


        Returns
        -------
        The AssociationRules object, ready to be used for fitting and prediction

       	Examples
       	--------
       	>>> idadb = IdaDataBase("BLUDB-TEST")
       	>>> idadf = IdaDataFrame(idadb, "GROCERIES")
       	>>> arules = AssociationRules("ASSOCRULES_TEST")
       	>>> arules.fit(idadf, transaction_id = "TID", item_id = "SID")

        Notes
        -----
        Inner parameters of the model can be printed and modified by using 
        get_params and set_params. But we recommend creating a new 
        AssociationRules model instead of modifying it.

        """
        # Get set at fit step
        self._idadf = None
        self._idadb = None
        self._transaction_id = None
        self._item_id = None
        self.nametable = None
        self.namecol = None

        # Get set at predict step
        self.outtable = None
        self.type = None
        self.limit = None
        self.sort = None

        self.modelname = modelname
        self.minsupport = minsupport
        self.maxlen = maxlen
        self.maxheadlen = maxheadlen
        self.minconf = minconf



    def get_params(self):
        """
        Return the parameters of the Association Rules model.
        """
        params = dict()
        params['modelname'] = self.modelname
        params['minsupport'] = self.minsupport
        params['maxlen'] = self.maxlen
        params['maxheadlen'] = self.maxheadlen
        params['minconf'] = self.minconf

        params['nametable'] = self.nametable
        params['namecol'] = self.namecol

        params['outtable'] = self.outtable
        params['type'] = self.type
        params['limit'] = self.limit
        params['sort'] = self.sort
        return params

    def set_params(self, **params):
        """
        Modify the parameters of the Association Rules model.
        """
        if not params:
            # Simple optimisation to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()
        for key, value in six.iteritems(params):
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s' %
                                     (key, self.__class__.__name__))
            setattr(self, key, value)
        return self

    def fit(self, idadf, transaction_id, item_id,  nametable=None, namecol=None, verbose=False):
        """
        Create an Association Rules model from an IdaDataFrame.

        Parameters
        ----------
        idadf : IdaDataFrame
            The IdaDataFrame to be used as input.
        transaction_id : str
            The column of the input table that identifies the transaction ID.
        item_id : str
            The column of the input table that identifies an item of the transaction.
        nametable : str, optional
            The IdaDataFrame or the name of the table that contains a mapping of the items in the input table
            and their names. The table name should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            The name table must contain at least two columns, where

            * The first column has the same name as the column that is
              contained in the item parameter of the input table
            * The second column has the same name as the name that is
              defined in the namecol parameter
        namecol : str, optional
            The column that contains the item name that is defined in the
            nametable parameter. You cannot specify this parameter if the
            nametable parameter is not specified.
        """
        if not type(idadf).__name__ == 'IdaDataFrame':
            raise TypeError("Argument should be an IdaDataFrame")
        if transaction_id not in idadf.columns:
            raise ValueError("transaction_id is not a column in " + idadf.name)
        if item_id is None:
            raise ValueError("item_id cannot be None (should be a column in " + idadf.name + ")")
        if item_id not in idadf.columns:
            raise ValueError("item_id is not a column in " + idadf.name)

        idadf._idadb._check_procedure("ASSOCRULES", "Association Rules")

        # Check the ID
        if transaction_id not in idadf.columns:
            raise ValueError("Transaction id column"+ transaction_id +" is not available in IdaDataFrame:" )

        self._idadb = idadf._idadb
        self._idadf = idadf
        self._transaction_id = transaction_id
        self._item_id = item_id

        if not (nametable is None or isinstance(nametable, ibmdbpy4nps.frame.IdaDataFrame)):
            nametable = ibmdbpy4nps.utils.check_tablename(nametable)
        self.nametable = nametable
        self.namecol = namecol

        # Check or create a model name
        if self.modelname is None:
            self.modelname = idadf._idadb._get_valid_modelname('ASSOCRULES_')
        else:
            self.modelname = ibmdbpy4nps.utils.check_tablename(self.modelname)
            if idadf._idadb.exists_model(self.modelname):
                idadf._idadb.drop_model(self.modelname)

        # Create a temporay view
        tmp_view_name = idadf.internal_state._create_view()
        # tmp_view_name = idadf.internal_state.current_state
        print(tmp_view_name)
        
        #if "." in tmp_view_name:
            #tmp_view_name = tmp_view_name.split('.')[-1]

        try:
            idadf._idadb._call_stored_procedure("ASSOCRULES ",
                                                 model = self.modelname,
                                                 intable = tmp_view_name,
                                                 tid = transaction_id,
                                                 item = item_id,
                                                 minsupport = self.minsupport,
                                                 maxlen = self.maxlen,
                                                 maxheadlen = self.maxheadlen,
                                                 minconf = self.minconf,
                                                 nametable = self.nametable,
                                                 namecol = self.namecol)
        except:
            raise
        finally:
            idadf.internal_state._delete_view()
            idadf.commit()

        self._retrieve_AssociationRules_Model(self.modelname, verbose)
        return

    def prune(self, itemsin = None, itemsout = None, minlen = 1, maxlen = None,
              minsupport = 0, maxsupport = 1, minlift = None, maxlift = None,
              minconf = None, maxconf = None, reset = False):
        """
        Prune the rules and patterns of an association rules model. To remove 
        rules and pattern which you are not interested in, you can use filters 
        to exclude these rules and patterns. These rules and patterns are then 
        marked as not valid in the model and are no longer shown.


        Parameters
        ----------
        itemsin : str or list, optional
            A list of item names that must be contained in the rules or
            patterns to be kept. The items are separated by semicolons. At
            least one of the listed items must be contained in a rule or
            pattern to be kept.
            For rules, the following conditions apply:

                * To indicate that the item must be contained in the head of then rule,
                  the item names can be succeeded by :h or :head.
                * To indicate that the item must be contained in the body of the rule,
                  the item names can be succeeded by :b or :body

            If this parameter is not specified, no constraint is applied.

        itemsout : str or list, optional
            A list of item names that must not be contained in the rules or
            patterns to be kept. The items are separated by semicolons.
            If this parameter is not specified, no constraint is applied.

        minlen : int, optional, >=1, default: 1
            The minimum number of items that are to be kept in the rules or
            patterns.

        maxlen : int, optional, >=1, default: the longest pattern of the model
            The maximum number of items that are to be kept in the rules or
            patterns.

        minsupport : float, optional, >=0.0 and <=maxsupport, default : 0
            The minimum support for the rules or patterns that are to be kept.

        maxsupport : float, optional, >=minsupport and <=1.0, default : 1
            The maximum support for the rules or patterns that are to be kept.

        minlift : float, optional, >=0.0 and <=maxlift, defaukt : 0
            The minimum lift of the rules or patterns that are to be kept.

        maxlift : float, optional, >=minlift, default: the maximum lift of the patterns of the model
            The maximum lift of the rules or patterns that are to be kept.

        minconf : float, optional, >=0.0 and <= maxconf, default : 0
            The minimum confidence of the rules that are to be kept.

        maxconf : float, optional, >=minconf and <= 1.0, default : 1
            The maximum confidence of the rules that are to be kept.

        reset : bool, optional, default: false
            If you specify reset=true, all rules and patterns are first reset
            to not pruned.
            If you specify reset=true or reset=false, the rules and patterns
            that are not to be kept are marked as pruned.
        """

        self._idadf._idadb._check_procedure("PRUNE_ASSOCRULES", "Pruning for Association Rules")

        if isinstance(itemsin, list):
            itemsin = ";".join(itemsin)

        try:
            self._idadf._idadb._call_stored_procedure("PRUNE_ASSOCRULES ",
                                                 model = self.modelname,
                                                 itemsin = itemsin,
                                                 itemsout = itemsout,
                                                 minlen = minlen,
                                                 maxlen = maxlen,
                                                 minsupport = minsupport,
                                                 maxsupport = maxsupport,
                                                 minlift = minlift,
                                                 maxlift = maxlift,
                                                 minconf = minconf,
                                                 maxconf = maxconf,
                                                 reset = reset)
        except:
            raise
        else:
        	return

    def predict(self, idadf, outtable=None, transaction_id=None, item_id=None,
                type="rules", limit=1, sort=None):
        """
        Apply the rules and patterns of an association rules model to other
        transactions. You can apply all rules or only specific rules according
        to specified criteria.

        Parameters
        ----------
        idadf : IdaDataFrame
            IdaDataFrame to be used as input.

        outtable : str, optional
            The name of the output table in which the mapping between the input 
            sequences and the associated rules or patterns is written.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If the parameter corresponds to an existing table in the database, it is
            replaced.

        transaction_id : str, optional
            The column of the input table that identifies the transaction ID. 
            By default, this is the same tid column that is specified in the 
            stored procedure to build the model.


        item_id : str, optional
            The column of the input table that identifies an item of the 
            transaction. By default, this is the same item column that is 
            specified in the stored procedure to build the model.


        type : str, optional, default : "rules"
            The type of information that is written in the output table. The 
            following values are possible: ‘rules’ and ‘patterns’.

        limit : int, optional, >=1, default: 1
            The maximum number of rules or patterns that is written in the 
            output table for each input sequence.

        sort : str or list, optional
            A list of keywords that indicates the order in which the rules or 
            patterns are written in the output table. The order of the list is 
            descending. The items are separated by semicolons. The following 
            values are possible: ‘support’, ‘confidence’, ‘lift’, and ‘length’. 
            The ‘confidence’ value can only be specified if the type parameter 
            is ‘rules’. If the type parameter is ‘rules’, the default is: 
            support;confidence;length.  If the type parameter is ‘patterns’, 
            the default is: support;lift;length. 

        Notes
        -----
        When "type" is set to "rules", it looks like nothing is returned.
        """
        if not isinstance(idadf, ibmdbpy4nps.IdaDataFrame):
            raise TypeError("Argument should be an IdaDataFrame")

        if sort is not None:
            sort = ';'.join(sort)

        if transaction_id is None:
            transaction_id = self.transaction_id
        if item_id is None:
            item_id = self.item_id

        # Check the ID
        if transaction_id not in idadf.columns:
            raise ValueError("Transaction id column"+ transaction_id +" is not available in IdaDataFrame." )

        if self._idadb is None:
            raise IdaAssociationRulesError("No Association rules model was trained before.")

        # The version where we don't replace the outtable if it exists but raise an exception
        #if outtable is not None:
        #    if idadf._idadb.exists_table(outtable):
        #        raise ValueError("Table "+ outtable +" already exists.")
        #else:
        #    outtable = idadf._idadb._get_valid_modelname('PREDICT_ASSOCRULES_')

        if self.outtable is None:
            self.outtable = idadf._idadb._get_valid_tablename('ASSOCRULES_')
        else:
            self.outtable = ibmdbpy4nps.utils.check_tablename(self.outtable)
            if idadf._idadb.exists_table(self.outtable):
                idadf._idadb.drop_table(self.outtable)

        self.outtable = outtable
        self.type = type
        self.limit = limit
        self.sort = sort

        # Create a temporay view
        tmp_view_name = idadf.internal_state._create_view()
        # tmp_view_name = idadf.internal_state.current_state
        
        #if "." in tmp_view_name:
            #tmp_view_name = tmp_view_name.split('.')[-1]

        try:
            idadf._idadb._call_stored_procedure("PREDICT_ASSOCRULES ",
                                                 model = self.modelname,
                                                 intable = tmp_view_name,
                                                 outtable = outtable,
                                                 tid = transaction_id,
                                                 item = item_id,
                                                 type = type,
                                                 limit = limit,
                                                 sort = sort
                                                 )
        except:
            raise
        finally:
            idadf.internal_state._delete_view()
            idadf.commit()

        self.labels_ = ibmdbpy4nps.IdaDataFrame(idadf._idadb, outtable)
        return self.labels_

    def fit_predict(self, idadf, transaction_id, item_id,  nametable=None,
    				namecol=None, outtable=None, type="rules", limit=1,
                    sort=None, verbose=False):
        """
        Convenience function for fitting the model and using it to make 
        predictions about the same dataset. See the fit and predict 
        documentation for an explanation about their attributes.


        Notes
        -----
        If you use this function, you are not able to use the prune step 
        between the fit and the predict step. However, you can still prune 
        afterwards and reuse the predict function.
        """
        #TODO: Is it relevant ? Result of predict on the same dataset looks empty
        #for type "rules"
        self.fit(idadf, transaction_id, item_id,  nametable, namecol, verbose)
        return self.predict(idadf, outtable, transaction_id, item_id, type, limit, sort)

    def describe(self, detail=False):
        """
        Parameters
        ----------
        detail: bool, optional. False by default.
        
        Returns
        -------
        Returns a description of Association Rules Model. If detail is set to False, then
        only a table with the summary of the rules is displayed, if detail is set to True,
        then thenintermediary tables are shown too.
        """
        if self._idadb is None:
            return self.get_params
        else:
            try:
                # res = self._idadb.ida_query("CALL IDAX.PRINT_MODEL('model = " + self.modelname +"')")
                res = self._idadb._call_stored_procedure("PRINT_MODEL ", model = self.modelname)
                if detail:
                    self._retrieve_AssociationRules_Model(self.modelname, verbose=True)
            except:
                raise
            else:
                print('Summary of the rules')
                print(res)
            return

    def _retrieve_AssociationRules_Model(self, modelname, verbose=False):
        """
        Retrieve information about the model to print the results. The 
        Association Rules IDAX/INZA function stores its result in 4 tables:

            * <MODELNAME>_ASSOCPATTERNS
            * <MODELNAME>_ASSOCPATTERNS_STATISTICS
            * <MODELNAME>_ASSOCRULES
            * <MODELNAME>_ITEMS

        Parameters
        ----------
        modelname : str
            The name of the model that is retrieved.
        verbose : bool, default: False
            Verbosity mode.

        Notes
        -----
        Needs better formatting instead of printing the tables
        """
        modelname = ibmdbpy4nps.utils.check_modelname(modelname)

        if self._idadb is None:
            raise IdaAssociationRulesError("No Association rules model was trained before.")

        # Note: The name of the columns in hardcoded, this is done so as a 
        # workaround for some bug in a specific ODBC linux driver. 
        # In case the implementation of the IDA method changes, this may break
        # But still would not be difficult to fix 

        assocpatterns = self._idadb.ida_query('SELECT * FROM ' + modelname + '_ASSOCPATTERNS')
        assocpatterns.columns = ["ITEMSETID","ITEMID"]
        assocpatterns.columns = [x.upper() for x in assocpatterns.columns]

        assocpatterns_stats = self._idadb.ida_query('SELECT * FROM ' + modelname + '_ASSOCPATTERNS_STATISTICS')
        assocpatterns_stats.columns = ["ITEMSETID" , "LENGTH" , "COUNT"  , "SUPPORT" , "LIFT"  ,"PRUNED"]
        assocpatterns_stats.columns = [x.upper() for x in assocpatterns_stats.columns]

        assocrules = self._idadb.ida_query('SELECT * FROM ' + modelname + '_ASSOCRULES')
        assocrules.columns = ["RULEID", "ITEMSETID", "BODYID", "HEADID", "CONFIDENCE", "PRUNED"]
        assocrules.columns = [x.upper() for x in assocrules.columns]

        items = self._idadb.ida_query('SELECT * FROM ' + modelname + '_ITEMS')
        items.columns = ["ITEMID","ITEM","ITEMNAME","COUNT","SUPPORT"]
        items.columns = [x.upper() for x in items.columns]

        if verbose is True:
            print("assocpatterns")
            print(assocpatterns)
            print(" ")

            print("assocpatterns_stats")
            print(assocpatterns_stats)
            print(" ")

            print("assocrules")
            print(assocrules)
            print(" ")

            print("items")
            print(items)
            print(" ")

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
idaDataFrame
---------
An efficient 2D container that looks like a panda's DataFrame and behave the
same, the only difference is that it uses only a reference to a remote database
instead of having the data loaded into memory

Also similar to its R counterpart, data.frame, except providing automatic data
alignment and a host of useful data manipulation methods having to do with the
labeling information
"""

# Ensure Python 2 compatibility
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import dict
from builtins import zip
from builtins import str
from builtins import int
from future import standard_library
standard_library.install_aliases()

import sys
import os
from copy import deepcopy
import warnings
from numbers import Number
from collections import OrderedDict

import numpy as np
import pandas as pd
from pandas.core.index import Index, RangeIndex

from lazy import lazy
import six

import ibmdbpy4nps
import ibmdbpy4nps.statistics
import ibmdbpy4nps.indexing
import ibmdbpy4nps.aggregation
import ibmdbpy4nps.filtering
import ibmdbpy4nps.utils

from ibmdbpy4nps.utils import timed, chunklist
from ibmdbpy4nps.internals import InternalState
from ibmdbpy4nps.exceptions import IdaDataFrameError
from ibmdbpy4nps.internals import idadf_state


class IdaDataFrame(object):
    """
    An IdaDataFrame object is a reference to a table in a remote Db2 Warehouse
    database. IDA stands for In-DataBase Analytics. IdaDataFrame copies the
    Pandas interface for DataFrame objects to ensure intuitive interaction for
    end-users.

    Examples
    --------
    >>> idadb = IdaDataBase('BLUDB') # See documentation for IdaDataBase
    >>> ida_iris = IdaDataFrame(idadb, 'IRIS')
    >>> ida_iris.cov()
                      sepal_length  sepal_width  petal_length  petal_width
    sepal_length      0.685694    -0.042434      1.274315     0.516271
    sepal_width      -0.042434     0.189979     -0.329656    -0.121639
    petal_length      1.274315    -0.329656      3.116278     1.295609
    petal_width       0.516271    -0.121639      1.295609     0.581006
    """

    # TODO: Check if everything is ok when selecting AND projecting with loc
    # TODO: BUG: Filtering after selection/projection

    def __init__(self, idadb, tablename, indexer = None):
        """
        Constructor for IdaDataFrame objects.

        Parameters
        ----------
        idadb : IdaDataBase
            IdaDataBase instance which contains the connection to be used.
        tablename : str
            Name of the table to be opened in the database.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
        indexer : str, optional
            Name of the column that should be used as an index. This is
            optional. However, if no indexer is given, the order of rows issued
            by the head and tail functions is not guaranteed. Also, several
            in-database machine learning algorithms need an indexer as a
            parameter to be executed.

        Attributes
        ----------
        _idadb : IdaDataBase
            IdaDataBase object parent of the current instance.
        tablename : str
            Name of the table self references.
        name : str
            Full name of the table self references, including schema.
        schema : str
            Name of the schema the table belongs to.
        indexer : str
            Name of the column used as an index. "None" if no indexer.
        loc : str
            Indexer that enables the selection and projection of IdaDataFrame
            instances. For more information, see the loc class documentation.
        internal_state : InternalState
            Object used to internally store the state of the IdaDataFrame. It
            also allows several non-destructive manipulation methods.
        type : str
            Type of the IdaDataFrame : “Table”, “View”, or “Unknown”.
        dtypes : DataFrame
            Data type in the database for each column.
        index : pandas.core.index
            Index containing the row names in this table.
        columns : pandas.core.index
            Index containing the columns names in this table.
        axes : list
            List containing columns and index attributes.
        shape : Tuple
            Number of rows and number of columns.

        Notes
        -----
        Attributes "type", "dtypes", "index", "columns", "axes", and "shape"
        are evaluated in a lazy way to avoid an overhead when creating an
        IdaDataFrame. Sometimes the index may be too big to be downloaded.

        Examples
        --------
        >>> idadb = IdaDataBase('BLUDB')
        >>> ida_iris = IdaDataFrame(idadb, "IRIS")
        """
        #TODO: Implement equality comparision between two IdaDataFrames

        if not idadb.__class__.__name__ == "IdaDataBase":
            idadb_class = idadb.__class__.__name__
            raise TypeError("Argument 'idadb' is of type %s, expected : IdaDataBase"%idadb_class)
        tablename = ibmdbpy4nps.utils.check_tablename(tablename)

        #idadb._reset_attributes("cache_show_tables")

        # TODO: Test what kind of error append when a table in use and cached
        # is suddently deleted

        if idadb.exists_table_or_view(tablename) is False:
            # Try again after refreshing the cache
            idadb._reset_attributes("cache_show_tables")
            # Refresh the show table cache in parent IdaDataBase, because a table
            # could have been created by other mean / user and we have to make sure
            # the lookup is done and is actual.
            if idadb.exists_table_or_view(tablename) is False:
                raise NameError("Table %s does not exist in the database %s."
                                %(tablename, idadb.data_source_name))


        self._idadb = idadb
        self._indexer = indexer

        # Initialise indexer object
        self.loc = ibmdbpy4nps.indexing.Loc(self)

        if "." in tablename:
            self.schema = tablename.split('.')[0]
            #self.name = tablename.split('.')[-1]
            self._name = tablename
            self.tablename = tablename.split('.')[-1]
        else:
            self.schema = idadb.current_schema
            self._name = idadb.current_schema + '.' + tablename
            self.tablename = tablename

        # self._name is the original name, this is a "final" variable

        # Push a reference to itself in its parent IdaDataBase
        self._idadb._idadfs.append(self)
        # TODO : self.size

        # A cache for unique value of each column
        self._unique = dict()

###############################################################################
### Attributes & Metadata computation
###############################################################################

    @lazy
    def internal_state(self):
        """
         InternalState instances manage the state of an IdaDataFrame instance
         and allow several non-destructive data manipulation methods, such as
         the selection, projection, filtering, and aggregation of columns.

        """
        return InternalState(self)


    @property
    @idadf_state
    def name(self):
        return self.internal_state.current_state


    @property
    def indexer(self):
        """
        The indexer attribute refers to the name of a column that should be
        used to index the table. This makes sense because Db2 Warehouse is a
        column-based database, so row IDs do not make sense and are not
        deterministic. As a consequence, the only way to address a particular
        row is to refer to it by its index. If no indexer is provided, ibmdbpy
        still works but a correct row order is not guaranteed as far as the
        dataset is not sorted. Also, note that the indexer column is not taken
        into account in data mining algorithms.
        """
        if hasattr(self, "_indexer"):
            return self._indexer
        else:
            None    

        
    @indexer.setter
    def indexer(self, value):
        """
        Basic checks for indexer :
        * The column exists in the table.
        * All values are unique.
        """
        if value is None:
            return

        if value not in self.columns:
            raise IdaDataFrameError("'%s' cannot be used as indexer "%value +
                                    " because this is not a column in '%s'"%self._name)

        del self.columns
        #count = self[value].count_distinct() ## TODO: TO FIX, should return directly just a number
        count = self.levels(value)
        if count < self.shape[0]:
            raise IdaDataFrameError("'%s' cannot be used as indexer "%value +
                                    " because it contains non unique values.")
        self._indexer = value


    @lazy
    def type(self):
        """
        Type of self: 'Table', 'View'  or 'Unknown'.

        Returns
        -------
        str
            idaDataFrame type.

        Examples
        --------
        >>> ida_iris.type
        'Table'
        """
        return self._get_type()

    @lazy
    @idadf_state(force = True)
    def dtypes(self):
        """
        Data type in database for each column in self.

        Returns
        -------
        DataFrame
            In-Database type for each columns.

        Examples
        --------
        >>> ida_iris.dtypes
                     TYPENAME
        sepal_length   DOUBLE
        sepal_width    DOUBLE
        petal_length   DOUBLE
        petal_width    DOUBLE
        species       VARCHAR
        """
        #import pdb ; pdb.set_trace()
        return self._get_columns_dtypes()

    @lazy
    @idadf_state
    # to deprecate
    def index(self):
        """
        Index containing the row names in self.

        Returns
        -------
        Index

        Examples
        --------
        >>> ida_iris.index
        Int64Index([  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
            ...
            140, 141, 142, 143, 144, 145, 146, 147, 148, 149],
           dtype='int64', length=150)

        Notes
        -----
        Because indexes in a database can be only numeric, it is not that
        interesting for an IdaDataFrame but can still be useful sometimes. The
        function can break if the table is too large. Ask for the user’s
        approval before downloading an index which has more than 10000 values.
        """
        return self._get_index()

    @lazy
    def columns(self):
        """
        Index containing the column names in self.

        Returns
        -------
        Index

        Examples
        --------
        >>> ida_iris.columns
        Index(['sepal_length', 'sepal_width', 'petal_length', 'petal_width',
        'species'],
        dtype='object')
        """
        if hasattr(self, "internal_state"):
            self.internal_state._create_view()
            cols = self._get_columns()
            self.internal_state._delete_view()
            return cols
        else:
            return self._get_columns()

    @lazy
    @idadf_state
    # to deprecate (no index)
    def axes(self):
        """
        List containing IdaDataFrame.columns and IdaDataFrame.index attributes.

        Returns
        -------
        list
            List containing two indexes (indexes and column attributes).

        Examples
        --------
        >>> ida_iris.axes
        [Int64Index([  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
             ...
             140, 141, 142, 143, 144, 145, 146, 147, 148, 149],
            dtype='int64', length=150),
        Index(['sepal_length', 'sepal_width', 'petal_length', 'petal_width',
        'species'],
        dtype='object')]
        """
        return [self.index, self.columns]

    @lazy
    @idadf_state
    def shape(self):
        """
        Tuple containing number of rows and number of columns.

        Returns
        -------
        tuple

        Examples
        --------
        >>> ida_iris.shape
        (150, 5)
        """
        return self._get_shape()

    @property
    @idadf_state
    def empty(self):
        """
        Boolean that is True if the table is empty (no rows).

        Returns
        -------
        Boolean
        """
        if self.shape[0] == 0:
            return True
        else:
            return False

    def __len__(self):
        """
        Number of records.

        Returns
        -------
        int

        Examples
        --------
        >>> len(idadf)
        150
        """
        return self.shape[0]

    def __iter__(self):
        """
        Iterate over columns.
        """
        return iter(self.columns)

    def __getitem__(self, item):
        """
        Enable label based projection (selection of columns) in IdaDataFrames.

        Enable slice based selection of rows in IdaDataFrames.

        Enable row filtering.

        The syntax is similar to Pandas.

        Examples
        --------

        >>> idadf['col1'] # return an IdaSeries

        >>> idadf[['col1']] # return an IdaDataFrame with one column

        >>> idadf[['col1', 'col2', 'col3']] # return an IdaDataFrame with 3 columns

        >>> idadf[0:9] # Select the 10 first rows

        >>> idadf[idadf['col1'] = "test"]
        # select of rows for which attribute col1 is equal to "test"

        Notes
        -----
        The row order is not guaranteed if no indexer is given and the dataset
        is not sorted
        """

        if isinstance(item, ibmdbpy4nps.filtering.FilterQuery):
            newidadf = self._clone()
            newidadf.internal_state.update(item)
            newidadf._reset_attributes(["shape"])
        else:
            if isinstance(item, slice):
                return self.loc[item]
            if not (isinstance(item,six.string_types)|isinstance(item, list)):
                raise KeyError(item)
            if isinstance(item, six.string_types):
                # Case when only one column was selected
                if item not in self.columns:
                    raise KeyError(item)
                newidaseries = self._clone_as_serie(item)

                # Form the new columndict
                for column in list(newidaseries.internal_state.columndict):
                    if column != item:
                        del newidaseries.internal_state.columndict[column]
                newColumndict = newidaseries.internal_state.columndict

                # Erase attributes
                newidaseries._reset_attributes(["columns", "shape", "dtypes"])
                # Set columns and columndict attributes
                newidaseries.internal_state.columns = ["\"%s\""%col for col in item]
                newidaseries.internal_state.columndict = newColumndict
                # Update, i.e. appends an entry to internal_state._cumulative
                newidaseries.internal_state.update()

                # Performance improvement
                # avoid, caused wrong dtypes for the result
                newidaseries.dtypes = self.dtypes.loc[[item]]

                return newidaseries

            # Case of multiple columns
            not_a_column = [x for x in item if x not in self.columns]
            if not_a_column:
                raise KeyError("%s"%not_a_column)

            newidadf = self._clone()

            # Form the new columndict
            newColumndict = OrderedDict()
            for col in item:
                # Column name as key, its definition as value
                newColumndict[col] = self.internal_state.columndict[col]

            # Erase attributes
            newidadf._reset_attributes(["columns", "shape", "dtypes"])
            # Set columns and columndict attributes
            newidadf.internal_state.columns = ["\"%s\""%col for col in item]
            newidadf.internal_state.columndict = newColumndict
            # Update, i.e. appends an entry to internal_state._cumulative
            newidadf.internal_state.update()

            # Performance improvement
            # avoid, caused wrong dtypes for the result
            newidadf.dtypes = self.dtypes.loc[item]

        return newidadf

    def __setitem__(self, key, item):
        """
        Enable the creation and aggregation of columns.

        Examples
        --------

        >>> idadf['new'] = idadf['sepal_length'] * idadf['sepalwidth']
        # select a new column as the product of two existing columns

        >>> idadf['sepal_length'] = idadf['sepal_length'] / idadf['sepal_length'].mean()
        # modify an existing column
        """
        if not (isinstance(item, IdaDataFrame)):
            raise TypeError("Modifying columns is supported only using "+
                            "IdaDataFrames.")
        if isinstance(key, six.string_types):
            key = [key]
        if len(key) != len(item.columns):
                raise ValueError("Wrong number of items passed %s, placement implies %s"%(len(item.columns),len(key)))

        #form the new columndict
        for newname, oldname in zip(key, item.columns):
            self.internal_state.columndict[newname] = item.internal_state.columndict[oldname]
        newColumndict = self.internal_state.columndict

        #erase attributes
        self._reset_attributes(["columns", "shape", "dtypes"])
        #set columns and columndict attributes
        self.internal_state.columndict = newColumndict
        self.internal_state.columns = ["\"%s\""%col for col in newColumndict.keys()]
        #update, i.e. appends an entry to internal_state._cumulative
        self.internal_state.update()

        # Flush the "unique" cache
        for column in key:
            if column in self._unique:
                del self._unique[column]

    def __delitem__(self, item):
        """
        Enable non-destructive deletion of columns using a Pandas style syntax.
        This happens inplace, which means that the current IdaDataFrame is
        modified.

        Examples
        --------
        >>> idadf = IdaDataFrame(idadb, "IRIS")
        >>> idadf.columns
        Index(['sepal_length', 'sepal_width', 'petal_length', 'petal_width', 'species'], dtype='object')
        >>> del idadf['sepal_length']
        >>> idadf.columns
        Index(['sepal_width', 'petal_length', 'petal_width', 'species'], dtype='object')
        """
        if not (isinstance(item,six.string_types)):
            raise TypeError
        if item not in self.columns:
            raise KeyError(item)

        # Flush the "unique" cache
        if item in self._unique:
            del self._unique[item]

        self._idadb.delete_column(self, item, destructive = False)
        return

    def __enter__(self):
        """
        Allow the object to be used with a "with" statement
        """
        return self

    def __exit__(self):
        """
        Allow the object to be used with a "with" statement. Make sure that
        allow possible views related to the IdaDataFrame with be deleted when
        the object goes out of scope
        """
        while self.internal_state.viewstack:
            try :
                view = self.internal_state.viewstack.pop()
                # Just a last check to make sure not to drop user's db
                if view != self.tablename:
                    drop = "DROP VIEW \"%s\"" %view
                    self._prepare_and_execute(drop, autocommit = True)
            except: pass

    #We decided not to allow columns access idadf.columnname like this for now.
    #We could decide to allow it but for this we may have to switch all
    #existing attributes to private ("_" as first character) so to avoid
    #conflicts between attributes and columns because for example IdaDataFrame
    #has an attribute "name" -> if a column is labelled "name" this will make
    #it unavailable. Pandas do also poorly manage this issue, for example :
    #DataFrame attribute "size".

    #May be implemented in the future

    #def __getattr__(self, name):
    #    """
    #    After regular attribute access, try look up the name
    #    This allows simpler access to columns for interactive use.
    #    """
        # Note: obj.x will always call obj.__getattribute__('x') prior to
        # calling obj.__getattr__('x').
        #if hasattr(self, name):
        #    return object.__getattribute__(self,name)
        #else:
    #    if name not in self.__dict__["columns"]:
            #return self[name]
    #        return object.__getattribute__(self, name)
    #    raise AttributeError("'%s' object has no attribute '%s'" %
    #                         (type(self).__name__, name))

    #def __setattr__(self, name, value):
    #    if name not in self.__dict__["columns"]:
    #        self.__dict__[name] = value
    #    else:
    #        raise ValueError("It is not allowed to set the value of a column in an IdaDataFrame.")

    def __lt__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using "<".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "lt", value)

    def __le__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using "<=".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "le", value)

    def __eq__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using "==".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "eq", value)

    def __ne__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using "!=".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "ne", value)

    def __ge__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using ">=".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "ge", value)

    def __gt__(self, value):
        """
        ibmdbpy.filtering.FilterQuery object when comparing self using ">".
        """
        return ibmdbpy4nps.filtering.FilterQuery(self.columns, self._name, "gt", value)

    ################ Arithmetic operations

    def __add__(self, other):
        """
        Perform an addition between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] + 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "add", other)

    def __radd__(self, other):
        """
        Enable the reflexivity of the addtion operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] + 3

        >>> ida = 3 + idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "add", self, swap = True)

    def __div__(self, other):
        """
        Perform a division between self and another IdaDataFrame or number.

        When __future__.division is not in effect.

        Examples
        --------
        >>> ida = idadf['sepal_length'] / 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "div", other)

    def __rdiv__(self, other):
        """
        Enable the reflexivity of the division operation.

        When __future__.division is not in effect.

        Examples
        --------
        >>> ida = idadf['sepal_length'] / 3

        >>> ida = 3 / idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "div", self, swap = True)

    def __truediv__(self, other):
        """
        Perform a division between self and another IdaDataFrame or number.

        When __future__.division is in effect.

        Examples
        --------
        >>> ida = idadf['sepal_length'] / 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "div", other)

    def __rtruediv__(self, other):
        """
        Enable the reflexivity of the division operation.

        When __future__.division is in effect.

        Examples
        --------
        >>> ida = idadf['sepal_length'] / 3

        >>> ida = 3 / idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "div", self, swap = True)

    def __floordiv__(self,other):
        """
        Perform an integer division between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] // 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "floordiv", other)

    def __rfloordiv__(self, other):
        """
        Enable the reflexivity of the integer division operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] // 3

        >>> ida = 3 // idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "floordiv", self, swap = True)

    def __mod__(self,other):
        """
        Perform a modulo operation between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] % 3

        Notes
        -----
        Arithmetic operations make sense if self has only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "mod", other)

    def __rmod__(self, other):
        """
        Enable the reflexivity of the modulo operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] % 3

        >>> ida = 3 % idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "mod", self, swap = True)

    def __mul__(self,other):
        """
        Perform a multiplication between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] * 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "mul", other)

    def __rmul__(self, other):
        """
        Enable the reflexivity of the multiplication operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] % 3

        >>> ida = 3 % idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "mul", self, swap = True)

    def __neg__(self):
        """
        Calculate the absolute negative of all columns in self.

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        other = None
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "neg", other)

    def __rpos__(self,other):
        """
        Calculate the absolute positive. No operation required.
        """
        return self

    def __pow__(self,other):
        """
        Perform a power operation between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] ** 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "pow", other)

    def __rpow__(self, other):
        """
        Enable the reflexivity of the power operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] ** 3

        >>> ida = 3 ** idadf['sepal_length']

        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "pow", self, swap = True)

    def __sub__(self,other):
        """
        Perform a substraction between self and another IdaDataFrame or number.

        Examples
        --------
        >>> ida = idadf['sepal_length'] - 3

        Notes
        -----
        Arithmetic operations only make sense if self contains only numeric columns.
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(self, "sub", other)

    def __rsub__(self, other):
        """
        Enable the reflexivity of the substraction operation.

        Examples
        --------
        >>> ida = idadf['sepal_length'] - 3

        >>> ida = 3 - idadf['sepal_length']
        """
        self._combine_check(other)
        return ibmdbpy4nps.aggregation.aggregate_idadf(other, "sub", self, swap = True)

    #def __truediv__(self,value): ########
    #    pass

    #def __concat__(self,value): ####
    #    pass

    #def __contains__(self,value):
    #    pass

    # TODO: Do the inplace versions (with "i" in front)

###############################################################################
### Database Features
###############################################################################

    ##############################
    ## Delegation from IdaDataBase
    ##############################

    def exists(self):
        """
        Convenience function delegated from IdaDataBase.

        Check if the data still exists in the database.
        """
        return self._idadb.exists_table_or_view(self._name)

    def is_view(self):
        """
        Convenience function delegated from IdaDataBase.

        Check if the IdaDataFrame corresponds to a view in the database.
        """
        if self.type == 'View':
            return True
        else:
            return False

    def is_table(self):
        """
        Convenience function delegated from IdaDataBase.

        Check if the IdaDataFrame corresponds to a table in the database.
        """
        if self.type == 'Table':
            return True
        else:
            return False

    def get_primary_key(self):
        # TODO: What happen if the primary key is composed of several columns ?
        """
        Get the name of the primary key of self, if there is one. Otherwise,
        this function returns 0. This function may be deprecated in future
        versions because it is not very useful.
        """
        name = self.internal_state.current_state
        pk = NULL
        # on Netezza primary keys are not supported
        if not self._idadb._is_netezza_system():
            pk = self.ida_query("SELECT NAME FROM SYSIBM.SYSCOLUMNS WHERE TBNAME = '" +
                                name + "' AND KEYSEQ > 0 ORDER BY KEYSEQ ASC", first_row_only = True)
        if pk:
            return pk[0]
        else:
            return False

    # Should we maybe allow this only in IdaDataBase object ?
    #@timed
    def ida_query(self, query, silent = False, first_row_only = False, autocommit = False):
        """
        Convenience function delegated from IdaDataBase.

        Prepare, execute and format the result of a query in a data frame or in
        a tuple. See the IdaDataBase.ida_query documentation.
        """
        return self._idadb.ida_query(query, silent, first_row_only, autocommit)

    def ida_scalar_query(self, query, silent = False, autocommit = False):
        """
        Convenience function delegated from IdaDataBase.

        Prepare and execute a query and return only the first element as a
        string. See the IdaDataBase.ida_scalar_query documentation.

        """
        return self._idadb.ida_scalar_query(query, silent, autocommit)



###############################################################################
### Data Exploration
###############################################################################

    def print(self):
        print(self.internal_state.get_state())

    @idadf_state
    def head(self, nrow=5, sort=True):
        """
        Print the n first rows of the instance, n is set to 5 by default.

        Parameters
        ----------
        nrow : int > 0
            Number of rows to be included in the result.

        sort: default is True
            If set to True and no indexer is set the data will be
            sorted by the first numeric column or if no numeric column
            is available by the first column of the dataframe.
            If set to False and no indexer is set the row order is not
            guaranteed and can vary with each execution. For big tables
            this option might save query processing time.


        Returns
        -------
        DataFrame or Series
            The index of the corresponding row number and the columns are all
            columns of self. If the IdaDataFrame has only one column, it
            returns a Series.


        Examples
        --------
        >>> ida_iris.head()
           sepal_length  sepal_width  petal_length  petal_width species
        0           5.1          3.5           1.4          0.2  setosa
        1           4.9          3.0           1.4          0.2  setosa
        2           4.7          3.2           1.3          0.2  setosa
        3           4.6          3.1           1.5          0.2  setosa
        4           5.0          3.6           1.4          0.2  setosa
        """
        if (nrow < 1) | (not isinstance(nrow, int)):
            raise ValueError("Parameter nrow should be an int greater than 0.")
        else:
            name = self.internal_state.current_state
            order = self.internal_state.get_order()
            if not " ORDER BY " in self.internal_state.get_state():
                if (self.indexer is not None)&(self.indexer in self.columns):
                    order = " ORDER BY \"" + str(self.indexer) + "\" ASC"
                elif self.indexer is None:
                    if sort:
                        column = self.columns[0]
                        if self._get_numerical_columns():
                            column = self._get_numerical_columns()[0]
                        order = " ORDER BY \"" + column + "\" ASC"
                    else:
                        order = ''
            data = self.ida_query("SELECT * FROM %s %s LIMIT %s "%(name, order, nrow))

            if data.shape[0] != 0:
                # otherwise column sort order is reverted
                if not 'SELECT ' in name:
                    columns = self.columns
                    data.columns = columns
#                data = ibmdbpy.utils._convert_dtypes(self, data)
                if isinstance(self, ibmdbpy4nps.IdaSeries):
                    data = pd.Series(data)
            return data

    # TODO : There is a warning in anaconda when there are missing values -> why ?
    @idadf_state
    def tail(self, nrow=5, sort=True):
        """
        Print the n last rows of the instance, n is set to 5 by default.

        Parameters
        ----------
        nrow : int > 0
            The number of rows to be included in the result.

        sort: default is True
            If set to True and no indexer is set the data will be
            sorted by the first numeric column or if no numeric column
            is available by the first column of the dataframe.
            If set to False and no indexer is set the row order is not
            guaranteed and can vary with each execution. For big tables
            this option might save query processing time.

        Returns
        -------
        DataFrame
            The index of the corresponding row number and the columns are all
            columns of self.


        Examples
        --------
        >>> ida_iris.tail()
             sepal_length  sepal_width  petal_length  petal_width    species
        145           6.7          3.0           5.2          2.3  virginica
        146           6.3          2.5           5.0          1.9  virginica
        147           6.5          3.0           5.2          2.0  virginica
        148           6.2          3.4           5.4          2.3  virginica
        149           5.9          3.0           5.1          1.8  virginica
        """
        if (nrow < 1) | (not isinstance(nrow, int)):
            raise ValueError("Parameter nrow should be an int greater than 0.")
        else:
            column_string = '\"' + '\", \"'.join(self.columns) + '\"'

            name = self.internal_state.current_state

            if " ORDER BY " in self.internal_state.get_state():
                query = "SELECT * FROM %s LIMIT %s "%(name, nrow)
                data = self.ida_query(query)
                data.columns = self.columns
                data.set_index(data[self.indexer], inplace=True)
            else:
                if self._idadb._is_netezza_system():
                    order = "ORDER BY NULL"
                else:
                    order = ""
                if self.indexer:
                    sortkey = str(self.indexer)
                    order = "ORDER BY \"" + sortkey + "\""
                else:
                    if sort:
                        sortkey = self.columns[0]
                        if self._get_numerical_columns():
                            sortkey = self._get_numerical_columns()[0]
                        order = "ORDER BY \"" + sortkey + "\""
                query = ("SELECT * FROM (SELECT * FROM (SELECT " + column_string +
                         ", ((ROW_NUMBER() OVER(" + order +
                         "))-1) AS ROWNUMBER FROM " + name +
                         ") AS TEMP1 ORDER BY ROWNUMBER DESC LIMIT " + str(nrow) +
                         " ) AS TEMP2 ORDER BY ROWNUMBER ASC")
                data = self.ida_query(query)
                data.set_index(data.columns[-1], inplace=True)  # Set the index
                data.columns = self.columns

            # del data.index.name
            #            data = ibmdbpy.utils._convert_dtypes(self, data)
            if isinstance(self, ibmdbpy4nps.IdaSeries):
                data = pd.Series(data[data.columns[0]])
            return data

    @idadf_state
    def pivot_table(self, values=None, columns=None, max_entries=1000,
                    sort=None, factor_threshold=20, interactive=False,
                    aggfunc='count'):
        """
        Compute an aggregation function over all rows of each column that is
        specified as a value on the dataset. The result grouped by the columns
        defined in “columns”.

        Parameters
        ----------
        values: str or list or str optional
            List of columns on which “aggfunc” is computed.
        columns: str or list or str optional
            List of columns that is used as an index and by which the
            dataframe is grouped.
        max_entries: int, default=1000
            The maximum number of cells to be part of the output. By default,
            set to 1000.
        sort: str, optional
            Admissible values are: “alpha” and “factors”.
                * If “alpha”, the index of the output is sorted according to the alphabetical order.
                * If “factors”, the index of the output will be sorted according to increasing number of the distinct values.

            By default, the index will be sorted in the same order that is specified in “columns” argument.

        factor_threshold: int, default: 20
            Number of distinct values above which a categorical column should
            not be considered categorical anymore and under which a numerical
            column column should not be considered numerical anymore.
        interactive: bool
            If True, the user is asked if he wants to display the output, given
            its size.
        aggfunc: str
            Aggregation function to be computed on each column specified in the
            argument “values”. Admissible values are: “count”, “sum”, “avg”.
            This entry is not case-sensitive.

        Returns
        -------
        Pandas Series with Multi-index (columns)

        Examples
        --------
        >>> val = ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']
        >>> ida_iris.pivot_table(values= val, aggfunc="avg")
                      species
        sepal_length  setosa        5.006
                      versicolor    5.936
                      virginica     6.588
        sepal_width   setosa        3.428
                      versicolor    2.770
                      virginica     2.974
        petal_length  setosa        1.462
                      versicolor    4.260
                      virginica     5.552
        petal_width   setosa        0.246
                      versicolor    1.326
                      virginica     2.026
        dtype: float64
        """
        # TODO : Support index
        from ibmdbpy4nps.statistics import pivot_table
        return pivot_table(idadf=self, values=values, columns=columns,
                           max_entries=max_entries, sort=sort,
                           factor_threshold=factor_threshold,
                           interactive=interactive, aggfunc=aggfunc)

    @idadf_state
    def groupby(self, by):
        # TODO: create an IdadataFrame groupby
        raise NotImplementedError()

    @idadf_state
    def merge(self):
        raise NotImplementedError()

    @idadf_state
    def concat(self):
        raise NotImplementedError()

    @idadf_state
    def join(self):
        raise NotImplementedError()

    # TODO : implement NULL FIRST option
    @idadf_state
    def sort(self, columns=None, axis=0, ascending=True, inplace=False):
        """
        Sort the IdaDataFrame row wise or column wise.

        Parameters
        ----------
        columns : str or list of str
            Columns that should be used to sort the rows in the IdaDataFrame.
            If columns is set to None and axis to 0, then the IdaDataFrame
            columns are sorted in lexicographical order.
        axis : int (0/1)
             Axis that is sorted. 0 for sorting row wise, 1 for sorting column
             wise.
        ascending : bool, default: True
            Sorting order, True : ascending, False : descending
        inplace : bool, default: False
            The current object is modified or creates a modified copy. If
            False, the function creates a modified copy of the current
            dataframe. If True,  the function modifies the current dataframe.

        Raises
        ------
        ValueError
            * When sorting by column (column not None), the axis value must be 0 (rows).
            * A column does not belong to self.
            * The axis argument has a value other than 0 or 1.

        Notes
        -----
        If columns is set to None and axis to 0, this undoes all sorting the
        IdaDataFrame and returns the original sorting in the Db2 Warehouse
        database.

        No actual changes are made in Db2 Warehouse, only the querying changes.
        Everything is registered in an InternalState object. Changes can be
        observed by using  head and tail function.
        """
        if isinstance(columns, six.string_types):
            columns = [columns]

        # Sanitiy check
        if columns:
            if axis != 0:
                raise ValueError("When sorting by column, axis must be 0 (rows)")
            for column in columns:
                if column not in self.columns:
                    raise ValueError("Column "+column+" is not in "+self._name)
        if axis not in [0,1]:
            raise ValueError("Allowed values for axis is 0 (rows) or 1 (columns)")

        if inplace:
            idadf = self
        else:
            idadf = self._clone()

        if axis:
            # Sort the columns in ascending or descending lexicographic order
            idadf.internal_state.columndict = OrderedDict(sorted(idadf.internal_state.columndict.items(), reverse = not ascending))
            idadf.internal_state.update()
        else:
            if columns:
                idadf.internal_state.set_order(columns, ascending)
                idadf.internal_state.update()
            else:
                if isinstance(idadf, ibmdbpy4nps.IdaSeries):
                    idadf.internal_state.set_order([idadf.column], ascending)
                    idadf.internal_state.update()
                else:
                    idadf.internal_state.reset_order()

        if not inplace:
            return idadf

    @idadf_state
    def levels(self, columns = None):
        # TODO: Test, doc, name?
        """
        Return the numbers of distinct values
        """
        if columns is not None:
            if isinstance(columns, six.string_types):
                columns = [columns]
            for column in columns:
                message = ''
                if column not in self.columns:
                    message += "Column %s does not belong to current idadataframe. \n"%column
                if message:
                    raise ValueError(message)
        else:
            columns = self.columns

        agglist = []
        for column in columns:
            agglist.append("COUNT(DISTINCT \"%s\")"%column)

        aggstr = ",".join(agglist)

        query = "SELECT %s FROM %s"%(aggstr, self.name)
        levels_tuple = self.ida_query(query, first_row_only = True)

        if len(levels_tuple) == 1:
            return levels_tuple[0]

        result = pd.Series(levels_tuple)
        result.index = columns
        return result


    #@timed
    @idadf_state
    def count_groupby(self, columns = None, count_only = False, having = None):
        """
        Count the occurence of the values of a column or group of columns
        """
        # TODO: Document, test
        if columns is not None:
            if isinstance(columns, six.string_types):
                columns = [columns]
            for column in columns:
                message = ''
                if column not in self.columns:
                    message += "Column %s does not belong to current idadataframe. \n"
                if message:
                    raise ValueError(message)
        else:
            columns = list(self.columns)
        if having:
            if not isinstance(having, Number):
                raise TypeError("having argument should be a number")

        select_columnstr = "\"" + "\",\"".join(columns) + "\", COUNT(*)"
        if count_only:
            select_columnstr = "COUNT(*)"
        groupby_columnstr = "\"" + "\",\"".join(columns) + "\""
        if having:
            groupby_columnstr = groupby_columnstr + " HAVING count >= %s"%having
        data = self.ida_query("SELECT %s AS count FROM %s GROUP BY %s"%(select_columnstr,self.name,groupby_columnstr))

        data.columns = columns + ["count"]
        return data

    def mean_freq_of_instance(self, columns = None):
        """
        Return the average occurence of the values of a column or group of columns
        """
        # TODO: Document, test
        if columns is not None:
            if isinstance(columns, six.string_types):
                columns = [columns]
            for column in columns:
                message = ''
                if column not in self.columns:
                    message += "Column %s does not belong to current idadataframe. \n"
                if message:
                    raise ValueError(message)
        else:
            columns = list(self.columns)

        select_columnstr = "COUNT(*)"
        groupby_columnstr = "\"" + "\",\"".join(columns) + "\""

        subquery = "SELECT %s AS count FROM %s GROUP BY %s"%(select_columnstr,self.name,groupby_columnstr)
        data = self.ida_scalar_query("SELECT AVG(DISTINCT count) FROM (%s)"%subquery)

        return int(data)

#    @timed
    @idadf_state
    def unique(self, column):
        """
        Return the unique values of a column
        """
        # TODO: Document, test
        if column in self._unique:
            return self._unique[column]

        #name = self.internal_state.current_state

        if not isinstance(column, six.string_types):
            raise TypeError("column argument not of string type")

        if column not in self.columns:
            # idadf.name somewhat false in case of modification
            raise ValueError("Undefined column \"%s\" in table %s"%(column, self._name))

        result = self.ida_query("SELECT DISTINCT \"%s\" FROM %s"%(column, self.name))

        self._unique[column] = result
        return result

    # TODO: to implement
    @timed
    @idadf_state
    def info(self, buf=None):
        """Some information about current IdaDataFrame. NOTIMPLEMENTED"""
        # There is a lot more
        from pandas.core.format import _put_lines

        if buf is None:  # pragma: no cover
            buf = sys.stdout

        lines = []

        lines.append(str(type(self)))
        lines.append(self.index.summary())

        if len(self.columns) == 0:
            lines.append('Empty %s' % type(self).__name__)
            _put_lines(buf, lines)
            return

###############################################################################
### Descriptive statistics
###############################################################################

    # TODO: to implement for categorical attributes
    @timed
    @idadf_state
    def describe(self, percentiles=[0.25, 0.50, 0.75]):
        """
        A basic statistical summary about current IdaDataFrame. If at least one
        numerical column exists, the summary includes:

            * The count of non-missing values for each numerical column.
            * The mean for each numerical column.
            * The standart deviation for each numerical column.
            * The minimum and maximum for each numerical column.
            * A list of percentiles set by the user (default : the quartiles).

        Parameters
        ----------
        idadf : IdaDataFrame
        percentiles : Float or list of floats, default: [0.25, 0.50, 0.75].
            percentiles to be computed on numerical columns.
            All values in percentiles must be > 0  and < 1.

        Returns
        -------
        summary: DataFrame, where
            * Index is the name of the computed values.
            * Columns are either numerical or categorical columns of self.

        """
        from ibmdbpy4nps.statistics import describe
        return describe(idadf=self, percentiles=percentiles)


    @timed
    @idadf_state
    def summary(self):
        """
        A basic statistical summary about current IdaDataFrame.

        Parameters
        ----------
        idadf : IdaDataFrame


        Returns
        -------
        summary: DataFrame, where
            * Index is the name of the computed values.
            * Columns are either numerical or categorical columns of self.

        """
        from ibmdbpy4nps.statistics import summary
        return summary(idadf=self)

    @timed
    @idadf_state
    def cov(self):
        """
        Compute the covariance matrix, composed of covariance coefficients
        between all pairs of columns in self.
        It must have at least two numeric columns.

        Returns
        -------
        covariance matrix: DataFrame
            The axes are the columns of self and the values are the covariance
            coefficients.
        """
        from ibmdbpy4nps.statistics import cov
        return cov(idadf=self)

    @timed
    @idadf_state
    def corr(self, method="pearson", features=None, ignore_indexer=True):
        """
        Compute the correlation matrix, composed of correlation coefficients
        between all pairs of columns in self.
        It must have at least two numeric columns.

        Parameters
        ----------
        method : str, default: pearson
            Method to be used to compute the correlation. By default, compute
            the pearson correlation coefficient. The Spearman rank correlation
            is also available. Admissible values are: "pearson", "spearman".

        Returns
        -------
        correlation matrix: DataFrame
            The axes are the columns of self and the values are the correlation
            coefficients.

        Notes
        -----
        For the Spearman rank correlation, the ordinal rank of columns is
        computed. For performance reasons this is easier to compute than the
        fractional rank traditionally computed for the Spearman rank
        correlation method. This strategy has the property that the sum of the
        ranking numbers is the same as under ordinal ranking. We then apply
        the pearson correlation coefficient method to these ranks.
        """
        from ibmdbpy4nps.statistics import corr
        #return corr(idadf=self, features=features, ignore_indexer=ignore_indexer)
        return corr(idadf=self)

    # TODO: to implement
    @timed
    @idadf_state
    def corrwith(self, other):
        """
        Compute the correlation matrix, composed of correlation coefficients
        between the columns of self and the columns of another IdaDataFrame.

        Parameters
        ----------
        other : DataFrame

        Returns
        -------
        correlation matrix: DataFrame
            The columns are the columns of self and the index the columns
            of other. The values are the covariance coefficients.
        """
        raise NotImplementedError("TODO")

    # TODO: to implement
    @timed
    @idadf_state
    def mode(self):
        """
        Compute the most common value for each non numeric column self. NOTIMPLEMENTED

        Returns
        -------
        mode: Series
        """
        raise NotImplementedError("TODO")

    @timed
    @idadf_state
    def quantile(self, q=0.5):
        """
        Compute row wise quantiles for each numeric column.

        Parameters
        ----------
        q : float or array-like, default 0.5 (50% quantile)
            0 <= q <= 1, the quantile(s) to compute

        Returns
        -------
        quantiles: Series or DataFrame
            If q is an array, the function returns a DataFrame in which the
            index is q. The columns are the columns of sel, and the values are
            the quantiles. If q is a float, a Series is returned where the
            index is the columns of self and the values are the quantiles.

        """
        from ibmdbpy4nps.statistics import quantile
        return quantile(idadf=self, q=q)

    @timed
    @idadf_state
    def mad(self):
        """
        Compute the mean absolute distance for all numeric columns of self.
        It must have at least two numeric columns.

        Returns
        -------
        mad: Series
            The index consists of the columns of self and the values are the mean absolute distance.
        """
        from ibmdbpy4nps.statistics import mad
        return mad(idadf=self)

    @timed
    @idadf_state
    def min(self):
        """
        Compute the minimum value for all numerics column of self.

        Returns
        -------
        min: Series
            The index consists of the columns of self and the values are the minimum.
        """
        from ibmdbpy4nps.statistics import ida_min
        return ida_min(idadf=self)

    @timed
    @idadf_state
    def max(self):
        """
        Compute the maximum value over for all numeric columns of self.

        Returns
        -------
        max: Series.
            The index consists of the columns of self and the values are the maximum.
        """
        from ibmdbpy4nps.statistics import ida_max
        return ida_max(idadf=self)

    @timed
    @idadf_state
    def count(self):
        """
        Compute the count of non-missing values for all columns of self.

        Returns
        -------
        count: Series.
            The index consists of the columns of self and the values are the number of non-missing values.
        """
        from ibmdbpy4nps.statistics import count
        return count(idadf=self)

    @timed
    @idadf_state
    def count_distinct(self):
        # deprecated, use levels instead
        """
        Compute the count of distinct values for all numeric columns of self.

        Returns
        -------
        disctinct count: Series
            The index consists of the columns of self and values are the number of distinct values.
        """
        from ibmdbpy4nps.statistics import count_distinct
        return count_distinct(idadf=self)

    @timed
    @idadf_state
    def std(self):
        """
        Compute the standard deviation for all numeric columns of self.

        Returns
        -------
        std: Series
            The index consists of the columns of self and the values are the standard deviation.
        """
        from ibmdbpy4nps.statistics import std
        return std(idadf=self)

    @timed
    @idadf_state
    def within_class_var(self, target, features = None, ignore_indexer=True):
        if features is None:
            numerical_columns = self._get_numerical_columns()
            features = [x for x in numerical_columns if x != target]
        else:
            if isinstance(features, six.string_types):
                features = [features]

        if ignore_indexer is True:
            if self.indexer:
                if self.indexer in features:
                    features.remove(self.indexer)

        result = pd.Series()

        #C = self.levels(target)
        N = len(self)
        if self._idadb._is_netezza_system():
            power_function = "POW"
        else:
            power_function = "POWER"

        if len(features) < 5:

            avglist = ["AVG(\"%s\") as \"average%s\""%(feature, index) for index, feature in enumerate(features)]
            sumlist = ["SUM(CAST(%s(\"%s\" - \"average%s\", 2) as DOUBLE))"%(power_function, feature, index) for index, feature in enumerate(features)]

            avgstr = ", ".join(avglist)
            sumstr = ", ".join(sumlist)

            subquery = "(SELECT \"%s\", %s FROM %s GROUP BY \"%s\") AS B"%(target, avgstr, self.name, target)
            condition = "ON A.\"%s\" = B.\"%s\""%(target, target)
            groupby = "GROUP BY A.\"%s\""%target
            query = "SELECT %s FROM %s AS A INNER JOIN %s %s %s"%(sumstr, self.name, subquery, condition, groupby)

            classvar = self.ida_query(query)
            if len(features) == 1:
                classvar = pd.DataFrame(classvar)

            C = len(classvar)
            if N == C:
                N += 1


            classvar.columns = pd.Index(features)
            result = pd.Series()
            for attr in classvar:
                result[attr] = classvar[attr].sum()/(N -C)

        else:
            chunkgen = chunklist(features, 5)
            result = pd.Series()
            for chunk in chunkgen:
                avglist = ["AVG(\"%s\") as \"average%s\""%(feature, index) for index, feature in enumerate(chunk)]
                sumlist = ["SUM(CAST(%s(\"%s\" - \"average%s\", 2) as DOUBLE))"%(power_function, feature, index) for index, feature in enumerate(chunk)]

                avgstr = ", ".join(avglist)
                sumstr = ", ".join(sumlist)

                subquery = "(SELECT \"%s\", %s FROM %s GROUP BY \"%s\") AS B"%(target, avgstr, self.name, target)
                condition = "ON A.\"%s\" = B.\"%s\""%(target, target)
                groupby = "GROUP BY A.\"%s\""%target
                query = "SELECT %s FROM %s AS A INNER JOIN %s %s %s"%(sumstr, self.name, subquery, condition, groupby)

                classvar = self.ida_query(query)

                if len(chunk) == 1:
                    classvar = pd.DataFrame(classvar)

                C = len(classvar)
                if N == C:
                    N += 1

                classvar.columns = pd.Index(chunk)

                for attr in classvar:
                    result[attr] = classvar[attr].sum()/(N -C)

        return result

    @timed
    @idadf_state
    def within_class_std(self, target, features = None, ignore_indexer= True):
        return np.sqrt(self.within_class_var(target, features, ignore_indexer))

    @timed
    @idadf_state
    def var(self):
        """
        Compute the variance for all numeric columns of self.

        Returns
        -------
        var: Series
            The index consists of the columns of self and the values are the variance.
        """
        from ibmdbpy4nps.statistics import var
        return var(idadf=self)

    @timed
    @idadf_state
    def mean(self):
        """
        Compute the mean for each numeric columns of self.

        Returns
        -------
        mean: Series
            The index consists of the columns of self and the values are the mean.
        """
        from ibmdbpy4nps.statistics import mean
        return mean(idadf=self)

    @timed
    @idadf_state
    def mean_groupby(self, groupby, features = None):
        if features is None:
            features = [x for x in self.columns if x != groupby]
        else:
            if isinstance(features, six.string_types):
                features = [features]

        avglist = ["CAST(AVG(\"%s\") as FLOAT) AS \"%s_AVG\""%(feature, feature) for feature in features]
        avgstr = ", ".join(avglist)

        query = "SELECT \"%s\", %s FROM %s GROUP BY \"%s\""%(groupby, avgstr, self.name, groupby)
        result = self.ida_query(query)
        result = result.pivot_table(index = result.columns[0])
        result.columns = pd.Index(features)
        return result


    @timed
    @idadf_state
    def sum(self):
        """
        Compute the sum of values for all numeric columns of self.

        Returns
        -------
        sum: Series
            The index consists of the columns of self and the values are the sum.
        """
        # Behave like having the option "numeric only" to true
        # TODO: Implement the option
        from ibmdbpy4nps.statistics import ida_sum
        return ida_sum(idadf=self)

    @timed
    @idadf_state
    def median(self):
        """
        Compute the median for all numeric columns of self.

        Returns
        -------
        median: Series
            The index consists of the columns of self and the values are the median.
        """
        # Behave like having the option "numeric only" to true
        # TODO: Implement the option
        from ibmdbpy4nps.statistics import median
        return median(idadf=self)

    # Maybe not be possible
    @idadf_state
    def rank(self):
        """Compute the rank over all entries for all columns of self."""
        raise NotImplementedError("TODO")

    # TODO : cumsum, cummean, cummcountm cummax, cumprod

###############################################################################
### Save current IdaDataFrame to Db2 Warehouse as a table
###############################################################################

    # TODO: Should this function be in IdaDataBase ?
    # To my mind, it is more intuitive to let it here, but it is "destructive"
    @timed
    @idadf_state
    def save_as(self, tablename, clear_existing = False):
        """
        Save self as a table name in the remote database with the name
        tablename. This function might erase an existing table if tablename
        already exists and clear_existing is True.

        """
        # TODO: to test !
        if tablename == self.tablename:
            if clear_existing is False:
                raise ValueError("Cannot overwrite current IdaDataFrame if "+
                                " clear_existing option set to False.")
            message = "Table %s already exists."%(tablename)
            warnings.warn(message, UserWarning)
            question = "Are you sure that you want to overwrite %s"%(tablename)
            display_yes = ibmdbpy4nps.utils.query_yes_no(question)
            if not display_yes:
                return
            tempname = self._idadb._get_valid_tablename()
            self._prepare_and_execute("CREATE TABLE %s AS (SELECT * FROM %s) WITH DATA"%(tempname, tablename))
            try:
                self._idadb.drop_table(tablename)
            except:
                self._idadb.drop_view(tablename)

            newidadf = IdaDataFrame(self._idadb, tempname)
            self._idadb.rename(newidadf, tablename)
            self = newidadf

        if self._idadb.exists_table_or_view(tablename):
            if clear_existing:
                message = "Table %s already exists."%(tablename)
                warnings.warn(message, UserWarning)
                question = "Are you sure that you want to overwrite %s"%(tablename)
                display_yes = ibmdbpy4nps.utils.query_yes_no(question)
                if not display_yes:
                    return
                try:
                    self._idadb.drop_table(tablename)
                except:
                    self._idadb.drop_view(tablename)
            else:
                raise NameError(("%s already exists, choose a different name "+
                                "or use clear_existing option.")%tablename)

        name = self.internal_state.current_state

        self._prepare_and_execute("CREATE TABLE %s AS (SELECT * FROM %s) WITH DATA"%(tablename, name))

        # Reset the cache
        self._idadb._reset_attributes("cache_show_tables")

###############################################################################
### Import as DataFrame
###############################################################################

    @timed
    @idadf_state
    def as_dataframe(self):
        """
        Download and return an in-memory representation of the dataset as
        a Pandas DataFrame.

        Returns
        -------
        DataFrame
            Columns and records are the same as in self.


        Examples
        --------
        >>> iris = ida_iris.as_dataframe()
        >>> iris.head()
           sepal_length  sepal_width  petal_length  petal_width species
        0           5.1          3.5           1.4          0.2  setosa
        1           4.9          3.0           1.4          0.2  setosa
        2           4.7          3.2           1.3          0.2  setosa
        3           4.6          3.1           1.5          0.2  setosa
        4           5.0          3.6           1.4          0.2  setosa
        """
        if os.environ['VERBOSE'] == 'True':
            # We use an empirical estimation
            # Experimental results :
            # 1M row, 12 columns => 550 secs
            # 3.77M, 12 columns => 2256 secs
            raw_estimation = ((self.shape[0]*self.shape[1]) / 4000000) * 3
            if raw_estimation < 0.01:
                # There is a small offset, which is negligeable for higher measurements
                estimation = str(raw_estimation*60 + 0.717) + " seconds."
            elif raw_estimation > 60:
                estimation = str(raw_estimation/60) + " hours."
            else:
                estimation = "%s minutes."%raw_estimation
            print("Your IdaDataFrame has %s rows and %s columns." % (self.shape[0], self.shape[1]))
            print("Estimated download time : ~ " + estimation)
            if raw_estimation > 30:
                message = ("Estimated Download time is greater than" +
                           " %s minutes.")%raw_estimation
                warnings.warn(message, UserWarning)
                question = "Do you want to download the dataset?"
                display_yes = ibmdbpy4nps.utils.query_yes_no(question)
                if not display_yes:
                    return

        data = self.ida_query(self.internal_state.get_state())
        data.columns = self.columns
        data.name = self.tablename
        # Handle datatypes
#        data = ibmdbpy.utils._convert_dtypes(self, data)
        return data

###############################################################################
### Connection Management
###############################################################################

# Connection is a bit different, this is made to allow user to work on several
# IdaDataFrame and close them, without closing the connection. This is why
# function "close" is overwritten and a function reopen is provided

    # Not sure commit and rollback should be here ?

    def commit(self):
        """
        Commit operations in the database.

        Notes
        -----

        All changes that are made in the database after the last commit,
        including those in the child IdaDataFrames, are commited.

        If the environment variable ‘VERBOSE’ is set to True, the commit
        operations are notified in the console.
        """
        self._idadb.commit()

    def rollback(self):
        """
        Rollback operations in the database.

        Notes
        -----

        All changes that are made in the database after the last commit,
        including those in the child IdaDataFrames, are discarded.
        """
        self._idadb.rollback()


###############################################################################
### Private functions
###############################################################################

    def _clone(self):
        """
        Clone the actual object.
        """
        newida = IdaDataFrame(self._idadb, self._name, self.indexer)
        newida.columns = self.columns
        newida.dtypes = self.dtypes     # avoid recomputing it
        # otherwise risk of infinite loop between
        # idadf.columns and internalstate.columndict

        # This is not possible to use deepcopy on an IdaDataFrame object
        # because the reference to the parents IdaDataBase with the connection
        # object is not pickleable. As a consequence, we create a new
        # IdaDataFrame and copy all the relevant attributes

        newida.internal_state.name = deepcopy(self.internal_state.name)
        newida.internal_state.ascending = deepcopy(self.internal_state.ascending)
        #newida.internal_state.views = deepcopy(self.internal_state.views)
        newida.internal_state._views = deepcopy(self.internal_state._views)
        newida.internal_state._cumulative = deepcopy(self.internal_state._cumulative)
        newida.internal_state.order = deepcopy(self.internal_state.order)
        newida.internal_state.columndict = deepcopy(self.internal_state.columndict)
        return newida

    def _clone_as_serie(self, column):
        """
        Clone the actual object as an IdaSeries and select one of its columns.
        """
        newida = ibmdbpy4nps.IdaSeries(idadb = self._idadb, tablename = self._name,
                                       indexer = self.indexer, column = column)

        newida.internal_state.name = deepcopy(self.internal_state.name)
        newida.internal_state.ascending = deepcopy(self.internal_state.ascending)
        #newida.internal_state.views = deepcopy(self.internal_state.views)
        newida.internal_state._views = deepcopy(self.internal_state._views)
        newida.internal_state._cumulative = deepcopy(self.internal_state._cumulative)
        newida.internal_state.order = deepcopy(self.internal_state.order)
        newida.internal_state.columndict = deepcopy(self.internal_state.columndict)
        return newida

    def _get_type(self):
        """
        Type of the IdaDataFrame : "Table", "View" or "Unknown".
        """
        if self._idadb.is_table(self._name):
            return "Table"
        elif self._idadb.is_view(self._name):
            return "View"
        else:
            return "Unkown"

    def _get_columns(self):
        """
        Index containing a list of the columns in self.
        """
        tablename = self.internal_state.current_state
        if self._idadb._con_type == 'odbc':
            if '.' in tablename:
                schema = tablename.split('.')[-2]
                tablename = tablename.split('.')[-1]
            else:
                schema = self._idadb.current_schema
            columns = self._idadb._con.cursor().columns(table=tablename, schema=schema)
            columnlist = [column[3] for column in columns]
            return Index(columnlist)
        elif self._idadb._con_type == 'jdbc':
            cursor = self._idadb._con.cursor()
            cursor.execute("SELECT * FROM %s"%tablename)
            columnlist = [column[0] for column in cursor.description]
            return Index(columnlist)

    def _get_all_columns_in_table(self):
        """Get all columns that exists in the physical table."""
        return self._get_columns()

    # TODO: To deprecate
    def _get_index(self, force=False):
        """
        Index containing a list of the row names in self.
        """

        rows = self.shape[0]

        return RangeIndex(0, rows, 1)

    def _get_shape(self):
        """
        Tuple containing the number of rows and the number of columns in self.
        """
        name = self.internal_state.current_state

        nrow = self.ida_scalar_query("SELECT CAST(COUNT(*) AS INTEGER) FROM %s"%name)
        ncol = len(self.columns)
        return (nrow, ncol)

    def _get_columns_dtypes(self):
        """
        DataFrame containing the column names and database types in self.
        """
        name = self.internal_state.current_state

        # In case the name is composed the following way : SCHEMA.TABLENAME
        # We need to separate it to keep only the TABLENAME for this query.
        if '.' in name :
            name = name.split('.')[-1]

        if self._idadb._is_netezza_system():
            nz_hidden_columns =  "('ROWID', 'DATASLICEID', 'CREATEXID', 'DELETEXID', '_PAGEID', '_EXTENTID') "
            query_select = ("SELECT COLUMN_NAME AS COLNAME, " +
                            "CASE WHEN strpos(TYPE_NAME, '(') = 0 THEN TYPE_NAME " +
                            "ELSE substr(TYPE_NAME, 1, strpos(TYPE_NAME,'(')-1) END AS TYPENAME " +
                            "FROM _V_SYS_COLUMNS ")
            # hidden Netezza columns are not included in the set of columns of this data frame
            query_where1 = "WHERE TABLE_NAME = \'%s\' AND COLUMN_NAME NOT IN " + nz_hidden_columns
            query_where2 = "AND SCHEMA = \'%s\' "
            query_order_by = "ORDER BY ORDINAL_POSITION "
        else:
            query_select = "SELECT COLNAME, TYPENAME FROM SYSCAT.COLUMNS "
            query_where1 = "WHERE TABNAME = \'%s\' "
            query_where2 = "AND TABSCHEMA = \'%s\' "
            query_order_by = "ORDER BY COLNO"

        if name.find("TEMP_VIEW_") == 0:
            #When the column names are going to be retrieved from a temporary
            #view that was created with the definition of the current state of
            #the IdaDataFrame, the schema name cannot be assumed as the same of
            #the IdaDataFrame. Also mind that the name of the temporary view
            #is thought to be random enough to avoid collisions
            data = self.ida_query((query_select + query_where1 + query_order_by)%(name))
        else:
            data = self.ida_query((query_select + query_where1 + query_where2 + query_order_by)%(name, self.schema))

        # Workaround for some ODBC version which does not get the entire
        # string of the column name in the cursor descriptor.
        # This is hardcoded, so be careful
        data.columns = ["COLNAME", "TYPENAME"]
        data.columns = [x.upper() for x in data.columns]
        data.set_index(keys='COLNAME', inplace=True)
        # del data.index.name
        return data

    def _reset_attributes(self, attributes):
        """
        Delete an attribute of self to force its refreshing at the next call.
        """
        if isinstance(attributes, six.string_types):
            attributes = [attributes]

        # Special case : resetting columns
        if "columns" in attributes:
             try:
                 del self.internal_state.columndict
             except:
                 pass

        ibmdbpy4nps.utils._reset_attributes(self, attributes)


###############################################################################
### DB2 Warehouse to pandas type mapping
###############################################################################

    def _table_def(self, factor_threshold=None):
        """
        Classify columns in the idaDataFrame into 4 classes: CATEGORICAL,
        STRING, NUMERIC or NONE. Use the database data type and a
        user-threshold “factor_threshold”:

            * CATEGORICAL columns that have a number of distinct values that is greater
              than the factor_threshold should be considered a STRING.
            * NUMERIC columns that have a number of distinct values that is smaller or equal
              to the factor_threshold should be considered CATEGORICAL.

        Returns
        -------
        DataFrame

            * Index is the columns of self.
            * Column "FACTORS" contains the number of distinct values.
            * Column "VALTYPE" contains the resulting class.

        Examples
        --------
        >>> ida_iris._table_def()
                     TYPENAME FACTORS      VALTYPE
        sepal_length   DOUBLE      35      NUMERIC
        sepal_width    DOUBLE      23      NUMERIC
        petal_length   DOUBLE      43      NUMERIC
        petal_width    DOUBLE      22      NUMERIC
        species       VARCHAR       3  CATEGORICAL
        """
        # We don't want to change the value of the attribute
        data = deepcopy(self.dtypes)

        def _valtype_from_dbtype(tup):
            """
            Decides if a column should be considered categorical or numerical
            """
            categorical_attributes = ['VARCHAR', 'CHARACTER VARYING', 'CHARACTER', 'VARGRAPHIC',
                                      'GRAPHIC', 'CLOB']
            numerical_attributes = ['SMALLINT', 'INTEGER', 'BIGINT', 'REAL',
                                    'DOUBLE', 'DOUBLE PRECISION', 'FLOAT', 'DECIMAL', 'NUMERIC']
            if tup[0] in categorical_attributes:
                if factor_threshold is None:
                    return "CATEGORICAL"
                elif tup[1] <= factor_threshold:
                    return "CATEGORICAL"
                else:
                    return "STRING"
            elif tup[0] in numerical_attributes:
                if factor_threshold is None:
                    return "NUMERIC"
                elif tup[1] > factor_threshold:
                    return "NUMERIC"
                else:
                    return "CATEGORICAL"
            else:
                return "NONE"

        data['FACTORS'] = ibmdbpy4nps.statistics._count_level(self, data.index.values)
        data['VALTYPE'] = [_valtype_from_dbtype(x) for x in
                           data[['TYPENAME', 'FACTORS']].to_records(index=0)]
        return data

    def _get_numerical_columns(self):
        """
        Get the columns of self that are considered as numerical. Their data
        type in the database determines whether these columns are numerical.
        The following data types are considered numerical:

            SMALLINT, INTEGER, BIGINT, REAL, DOUBLE, FLOAT, DECIMAL, NUMERIC.

        Returns
        -------
        list
            List of numerical column names.

        Examples
        --------
        >>> ida_iris._get_numerical_columns()
        ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']
        """
        num = ['SMALLINT', 'INTEGER', 'BIGINT', 'REAL',
                            'DOUBLE', 'DOUBLE PRECISION', 'FLOAT', 'DECIMAL', 'NUMERIC']
        return list(self.dtypes.loc[self.dtypes['TYPENAME'].isin(num)].index)

    def _get_categorical_columns(self):
        """
        Get the columns of self that are considered as categorical. Their data
        type in the database determines whether these columns are categorical.
        The following data types are considered categorical:

            VARCHAR,CHARACTER, VARGRAPHIC, GRAPHIC, CLOB.

        Returns
        -------
        list
            List of categorical column names.

        Examples
        --------
        >>> ida_iris._get_categorical_columns()
        ['species']
        """
        cat = ['VARCHAR', 'CHARACTER', 'VARGRAPHIC', 'GRAPHIC', 'CLOB']
        return list(self.dtypes.loc[self.dtypes['TYPENAME'].isin(cat)].index)

    def _prepare_and_execute(self, query, autocommit = True, silent = False):
        """
        Prepare and execute a query.

        Parameters
        ----------
        query : str
            Query to be executed.
        autocommit : bool
             If set to true, the autocommit function is available.
        """
        return self._idadb._prepare_and_execute(query, autocommit, silent)

    def _autocommit(self):
        """
        Autocommit the connection. If the environment variable ‘AUTOCOMMIT’ is
        set to True, the function commits the changes.

        Notes
        -----

        If you commit, all changes that are made in the database after the last
        commit, including those in the child IdaDataFrames, are commited.

        If the environment variable ‘VERBOSE’ is set to True, the autocommit
        operations are notified in the console.
        """
        self._idadb._autocommit()

    def _combine_check(self, other):
        """
        Check if self and other refer to the same table and if all columns in
        self and other are numeric. This sanity check is used before performing
        aggregation operations between IdaDataFrame/IdaSeries.
        """
        def check_numeric_columns(idaobject):
            not_valid = []
            numeric_columns = idaobject._get_numerical_columns()
            for column in idaobject.columns:
                if column not in numeric_columns:
                    not_valid.append(column)
            if not_valid:
                raise TypeError("Arithmetic operation are not defined for %s"%not_valid)

        if isinstance(other, IdaDataFrame) | isinstance(other, ibmdbpy4nps.IdaSeries):
            if self._name != other._name:
                raise IdaDataFrameError("It is not possible to aggregate columns using columns of a different table.")

        if not(isinstance(other, IdaDataFrame) | isinstance(other, ibmdbpy4nps.IdaSeries) | isinstance(other, Number)):
            if other is not None:
                raise TypeError("Aggregation makes only sense with numbers, "+
                                "or IdaDataFrames refering to the same table.")

        check_numeric_columns(self)
        if isinstance(other, ibmdbpy4nps.IdaSeries)|isinstance(other, IdaDataFrame):
            check_numeric_columns(other)

    
    def delete_na(self, columns, logic="any", inplace=False):
        """
        Filter rows containing NULL values. Can be done in a destructive way (rows are deleted in the physical table)
        or in a non destructive way (a new IdaDataFrame is defined, original table and IdaDataFrame are preserved). 
        
        Parameters
        -----------
        columns : list of strings
            list of eligible column names
        logic : str, optional
            "any" by default. 
            If logic is set to "any" then all rows which contain a NaN value in any 
            (id est at least one) of the cited columns is deleted, this is a union condition;
            if logic is set to "all", then only rows containing null values in all the fields are deleted, this is an intersection condition.         
        inplace : bool, optional 
            False by default. 
            If True, then the underlying table is physically modified,
            if False, then the original objects remain unaffected, a copy of the IdaDataFrame is made and modified. 
        
        Returns
        --------
        If inplace is True, the original table is modified and the IdaDataFrame will be modified accordingly. No return.
        If inplace is False, a new IdaDataFrame is returned, the original table is not affected. This new IdaDataFrame points to a 
        new table which has been created according to the user defined criterions on columns.
        
        Examples
        --------
        >>> idadf.delete_na(["COL_1", "COL_2"], inplace = True)
        >>> #each row of the original table is physically deleted if there is a NULL value in one of the listed columns
        >>> new_idadf = idadf.delete_na(["COL_1", "COL_2"], logic = 'all')
        >>> # a new IdaDataFrame is created, rows will be selected only if all the listed columns have a NULL value.
        """
        
        # Check if list columns is not empty
        if len(columns)<1:
            raise IdaDataFrameError("You must specify the columns as a list of eligible names")
        # Check if column names are eligible
        for column_name in columns:
            if not isinstance(column_name, six.string_types):
                raise TypeError("%s is not of string type"%(column_name))
            if column_name not in self.columns:
                raise ValueError("%s refers to a columns that doesn't exists in self"%(column_name))
        
        idadb = self._idadb
        
        if inplace == True:
            # Explain
            print("The table %s will be physically modified." %self.tablename)
            print("Any IdaDataFrame pointing this table might be modified accordingly.")
            # Write DELETE query on original table
            tablename = self.tablename
            query = 'DELETE FROM %s WHERE "%s" IS NULL'%(tablename, columns[0])
            
            if len(columns)>1:
                if logic == "any":
                    for col in columns[1:]:
                        query = query + ' OR "%s" IS NULL'%col
                if logic == "all":
                    for col in columns[1:]:
                        query = query + ' AND "%s" IS NULL'%col                        
            idadb.ida_query(query)

        else:            
            # SELECT statement on the original table to create a view
            tablename0 = self.tablename 
            query = 'SELECT * FROM %s WHERE "%s" IS NOT NULL'%(tablename0, columns[0])
            
            if len(columns)>1:
                if logic == "any":
                    for col in columns[1:]:
                        query = query + ' AND "%s" IS NOT NULL'%col
                if logic == "all":
                    for col in columns[1:]:
                        query = query + ' OR "%s" IS NOT NULL'%col                                               
            # Create view with this select statement
            viewname = idadb._get_valid_tablename(prefix="VIEW_")
            self._prepare_and_execute("CREATE VIEW " + viewname + " AS "+ query)
            # Initiate the modified table under a random name
            tablename = idadb._get_valid_tablename(prefix="DATA_FRAME_")
            idadb._prepare_and_execute("CREATE TABLE %s AS (SELECT * FROM %s) WITH DATA"%(tablename,viewname))
            print('A new table with filtered rows is available under the name %s.' %tablename)
            print('The newly created IdaDataFrame refers to this new table.')
            # Drop the view 
            idadb.drop_view(viewname)            
            # Define a new IdaDataFrame pointing to the new table
            idadf = IdaDataFrame(idadb, tablename, indexer = self.indexer)
            return idadf
    
            


            
        